"""Structured result helpers for runner outputs."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any


DEFAULT_FILENAME = "<memory>"


def normalize_filename(filename: str | None) -> str:
    return filename or DEFAULT_FILENAME


@dataclass
class Result:
    ok: bool
    stage: str
    filename: str
    stdout: str = ""
    stderr: str = ""
    result: Any = None
    errors: list[dict] = field(default_factory=list)
    diagnostics: list[dict] = field(default_factory=list)
    error: dict | None = None
    extras: dict[str, Any] = field(default_factory=dict)

    def to_dict(self) -> dict:
        payload = {
            "ok": bool(self.ok),
            "stage": self.stage,
            "filename": self.filename,
            "stdout": self.stdout,
            "stderr": self.stderr,
            "result": self.result,
            "errors": list(self.errors),
            "diagnostics": list(self.diagnostics),
            "error": self.error,
        }
        if self.extras:
            payload.update(self.extras)
        return payload

    @classmethod
    def success(
        cls,
        *,
        stage: str,
        filename: str | None = None,
        stdout: str = "",
        stderr: str = "",
        result: Any = None,
        diagnostics: list[dict] | None = None,
        extras: dict[str, Any] | None = None,
    ) -> "Result":
        return cls(
            ok=True,
            stage=stage,
            filename=normalize_filename(filename),
            stdout=stdout,
            stderr=stderr,
            result=result,
            errors=[],
            diagnostics=list(diagnostics or []),
            error=None,
            extras=dict(extras or {}),
        )

    @classmethod
    def failure(
        cls,
        *,
        stage: str,
        filename: str | None = None,
        stdout: str = "",
        stderr: str = "",
        result: Any = None,
        errors: list[dict] | None = None,
        diagnostics: list[dict] | None = None,
        error: dict | None = None,
        extras: dict[str, Any] | None = None,
    ) -> "Result":
        return cls(
            ok=False,
            stage=stage,
            filename=normalize_filename(filename),
            stdout=stdout,
            stderr=stderr,
            result=result,
            errors=list(errors or []),
            diagnostics=list(diagnostics or []),
            error=error,
            extras=dict(extras or {}),
        )



import unittest

from nodus.runtime.errors import BytecodeVersionError
from nodus.vm.vm import VM


class BytecodeVersionTests(unittest.TestCase):
    def test_bytecode_version_mismatch(self):
        with self.assertRaises(BytecodeVersionError):
            VM({"bytecode_version": 999, "instructions": []}, {}, code_locs=[])


if __name__ == "__main__":
    unittest.main()



import io
import unittest
from contextlib import redirect_stdout

import nodus as lang
from nodus.runtime.module_loader import ModuleLoader


def run_program(src: str, source_path: str | None = None):
    _loader = ModuleLoader(project_root=None)
    code, functions, code_locs = _loader.compile_only(
        src,
        module_name=source_path or "<memory>",
    )
    vm = lang.VM(code, functions, code_locs=code_locs, source_path=source_path)
    buf = io.StringIO()
    with redirect_stdout(buf):
        vm.run()
    return buf.getvalue().splitlines()


class VmErrorWrappingTests(unittest.TestCase):
    def test_python_type_error_wrapped(self):
        src = '"a" - 1'
        with self.assertRaises(lang.LangRuntimeError) as cm:
            run_program(src, source_path="main.nd")
        self.assertEqual(cm.exception.kind, "type")
        self.assertIsNotNone(cm.exception.line)
        self.assertIn("Cannot subtract", str(cm.exception))


if __name__ == "__main__":
    unittest.main()



"""Compatibility wrapper around Python's stdlib `types` plus Nodus types."""

import os as _os

_stdlib_types_path = _os.path.join(_os.path.dirname(_os.__file__), "types.py")
_stdlib_namespace: dict[str, object] = {"__file__": _stdlib_types_path, "__name__": "_stdlib_types"}
with open(_stdlib_types_path, "r", encoding="utf-8") as _f:
    exec(compile(_f.read(), _stdlib_types_path, "exec"), _stdlib_namespace)

for _name, _value in list(_stdlib_namespace.items()):
    if _name in {"__builtins__", "__doc__", "__file__", "__name__", "__package__", "__spec__"}:
        continue
    globals()[_name] = _value

from nodus.frontend.type_system import ANY, BOOL, FLOAT, FUNCTION, INT, LIST, NIL, RECORD, STRING, FunctionType, NodusType

__all__ = list(_stdlib_namespace.get("__all__", [])) + [
    "NodusType",
    "FunctionType",
    "ANY",
    "INT",
    "FLOAT",
    "STRING",
    "BOOL",
    "LIST",
    "RECORD",
    "FUNCTION",
    "NIL",
]



"""Runtime registry and dispatch for agent calls."""

from __future__ import annotations

from nodus.result import Result, normalize_filename
from nodus.vm.runtime_values import clone_json_value, is_json_safe, payload_keys


AGENT_REGISTRY: dict[str, dict] = {}


def register_agent(name: str, handler, *, description: str | None = None, payload_schema: dict | None = None) -> None:
    AGENT_REGISTRY[name] = {
        "handler": handler,
        "spec": {
            "name": name,
            "description": description or f"Agent handler for {name}",
            "parameters": payload_schema or {"type": "object"},
        },
    }


def unregister_agent(name: str) -> None:
    AGENT_REGISTRY.pop(name, None)


def available_agents() -> list[str]:
    return sorted(AGENT_REGISTRY.keys())


def describe_agent(name: str):
    if not isinstance(name, str):
        return None
    entry = AGENT_REGISTRY.get(name)
    if entry is None:
        return None
    return dict(entry["spec"])


def call_agent(name, payload, *, vm=None) -> dict:
    filename = normalize_filename(getattr(vm, "source_path", None))
    if not isinstance(name, str) or not name:
        return _agent_error("Agent name must be a non-empty string", filename, name=name)
    if not is_json_safe(payload):
        return _agent_error("Agent payload must be JSON-safe", filename, name=name)

    _emit(vm, "agent_call_start", name=name, payload=payload)
    entry = AGENT_REGISTRY.get(name)
    if entry is None:
        result = _agent_error(f"No handler registered for agent '{name}'", filename, name=name)
        _emit(vm, "agent_call_fail", name=name, payload=payload, ok=False, error=_error_message(result))
        return result

    try:
        handler_result = entry["handler"](clone_json_value(payload))
    except Exception as err:
        result = _agent_error(str(err), filename, name=name)
        _emit(vm, "agent_call_fail", name=name, payload=payload, ok=False, error=_error_message(result))
        return result

    if not is_json_safe(handler_result):
        result = _agent_error("Agent handler returned a non-serializable value", filename, name=name)
        _emit(vm, "agent_call_fail", name=name, payload=payload, ok=False, error=_error_message(result))
        return result

    result = Result.success(
        stage="agent_call",
        filename=filename,
        stdout="",
        stderr="",
        result=clone_json_value(handler_result),
    ).to_dict()
    _emit(vm, "agent_call_complete", name=name, payload=payload, ok=True)
    return result


def _agent_error(message: str, filename: str, *, name: str | None = None) -> dict:
    legacy = {"type": "agent", "message": message, "path": filename}
    if name is not None:
        legacy["agent"] = name
    return Result.failure(
        stage="agent_call",
        filename=filename,
        stdout="",
        stderr="",
        errors=[{"type": "AgentError", "message": message, "agent": name}],
        error=legacy,
    ).to_dict()


def _error_message(result: dict | None) -> str:
    if not isinstance(result, dict):
        return "Agent call failed"
    err = result.get("error")
    if isinstance(err, dict):
        return str(err.get("message", "Agent call failed"))
    return "Agent call failed"


def _emit(vm, event_type: str, *, name: str, payload, ok: bool | None = None, error: str | None = None) -> None:
    if vm is None or getattr(vm, "event_bus", None) is None:
        return
    data = {"payload_keys": payload_keys(payload)}
    if hasattr(vm, "runtime_adapter_event_data"):
        data.update(vm.runtime_adapter_event_data(payload, ok=ok, error=error))
    vm.event_bus.emit_event(event_type, name=name, data=data)



"""Minimal stdio LSP server for Nodus."""

from __future__ import annotations

import json
import os
import re
import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import BinaryIO
from urllib.parse import unquote, urlparse

from nodus.frontend.ast.ast_nodes import (
    Assign,
    Attr,
    Bin,
    Block,
    Call,
    ExprStmt,
    FnDef,
    FnExpr,
    For,
    ForEach,
    GoalDef,
    If,
    Import,
    Index,
    IndexAssign,
    Let,
    ListLit,
    MapLit,
    ModuleAlias,
    Nil,
    Num,
    Param,
    Print,
    RecordLiteral,
    Return,
    Str,
    Throw,
    TryCatch,
    Unary,
    Var,
    While,
    WorkflowDef,
    WorkflowStateDecl,
    Yield,
)
from nodus.frontend.lexer import KEYWORDS, Tok, tokenize
from nodus.frontend.parser import Parser
from nodus.frontend.type_system import ANY, BOOL, FLOAT, FUNCTION, INT, LIST, NIL, RECORD, STRING, FunctionType, combine_types
from nodus.runtime.dependency_graph import DependencyGraph
from nodus.tooling.diagnostics import WorkspaceDiagnosticEngine
from nodus.tooling.loader import collect_module_info, ensure_project_root, resolve_import_path


COMPLETION_KIND_TEXT = 1
COMPLETION_KIND_FUNCTION = 3
COMPLETION_KIND_VARIABLE = 6
COMPLETION_KIND_MODULE = 9
COMPLETION_KIND_KEYWORD = 14
IDENTIFIER_RE = re.compile(r"[A-Za-z_][A-Za-z0-9_]*$")


@dataclass
class DefinitionRecord:
    name: str
    kind: str
    path: str
    uri: str
    line: int
    col: int
    end_col: int
    detail: str
    type_text: str | None = None
    signature: str | None = None


@dataclass
class ReferenceRecord:
    name: str
    line: int
    col: int
    end_col: int
    target: DefinitionRecord | None = None


@dataclass
class ModuleRecord:
    path: str
    uri: str
    exports: dict[str, DefinitionRecord] = field(default_factory=dict)


@dataclass
class DocumentState:
    uri: str
    path: str
    text: str
    diagnostics: list[dict]
    tokens: list[Tok]
    definitions: list[DefinitionRecord]
    references: list[ReferenceRecord]
    visible_symbols: dict[str, DefinitionRecord]
    module_aliases: dict[str, ModuleRecord]


def _path_to_uri(path: str) -> str:
    return Path(path).resolve().as_uri()


def _uri_to_path(uri: str) -> str:
    parsed = urlparse(uri)
    if parsed.scheme != "file":
        return os.path.abspath(uri)
    path = unquote(parsed.path)
    if os.name == "nt" and path.startswith("/"):
        path = path[1:]
    return os.path.realpath(path)


def _read_message(stream: BinaryIO) -> dict | None:
    headers: dict[str, str] = {}
    while True:
        line = stream.readline()
        if not line:
            return None
        if line in {b"\r\n", b"\n"}:
            break
        decoded = line.decode("utf-8").strip()
        if ":" not in decoded:
            continue
        key, value = decoded.split(":", 1)
        headers[key.strip().lower()] = value.strip()
    length_text = headers.get("content-length")
    if not length_text:
        return None
    body = stream.read(int(length_text))
    if not body:
        return None
    return json.loads(body.decode("utf-8"))


def _write_message(stream: BinaryIO, payload: dict) -> None:
    body = json.dumps(payload, separators=(",", ":")).encode("utf-8")
    header = f"Content-Length: {len(body)}\r\n\r\n".encode("ascii")
    stream.write(header)
    stream.write(body)
    stream.flush()


def _lsp_position(line: int | None, col: int | None) -> dict:
    return {
        "line": max((line or 1) - 1, 0),
        "character": max((col or 1) - 1, 0),
    }


def _range_for(line: int | None, col: int | None, end_col: int | None = None) -> dict:
    start = _lsp_position(line, col)
    end = _lsp_position(line, end_col if end_col is not None else col)
    if end["character"] < start["character"]:
        end["character"] = start["character"]
    return {"start": start, "end": end}


def _type_to_text(type_value) -> str:
    if isinstance(type_value, FunctionType):
        params = ", ".join(_type_to_text(param) for param in type_value.params)
        return f"fn({params}) -> {_type_to_text(type_value.return_type)}"
    return type_value.name


def _name_to_type(name: str | None):
    if name == "int":
        return INT
    if name == "float":
        return FLOAT
    if name == "string":
        return STRING
    if name == "bool":
        return BOOL
    if name == "list":
        return LIST
    if name == "record":
        return RECORD
    if name == "function":
        return FUNCTION
    if name == "nil":
        return NIL
    return ANY


class _TypeEnv:
    def __init__(self) -> None:
        self.scopes: list[dict[str, object]] = [{}]

    def push(self) -> None:
        self.scopes.append({})

    def pop(self) -> None:
        self.scopes.pop()

    def bind(self, name: str, value) -> None:
        self.scopes[-1][name] = value

    def lookup(self, name: str):
        for scope in reversed(self.scopes):
            if name in scope:
                return scope[name]
        if name == "clock":
            return FunctionType([], FLOAT)
        if name in {"type", "str", "input"}:
            return FunctionType([ANY], STRING)
        if name == "len":
            return FunctionType([ANY], INT)
        if name == "print":
            return FunctionType([ANY], NIL)
        return ANY


def _identifier_column(lines: list[str], line: int, name: str, start_col: int) -> int:
    if line <= 0 or line > len(lines):
        return max(start_col, 1)
    text = lines[line - 1]
    search_from = max(start_col - 1, 0)
    pattern = re.compile(rf"\b{re.escape(name)}\b")
    match = pattern.search(text, search_from)
    if match is not None:
        return match.start() + 1
    fallback = text.find(name, search_from)
    return fallback + 1 if fallback >= 0 else max(start_col, 1)


def _infer_expr_type(expr, env: _TypeEnv):
    if isinstance(expr, Num):
        if expr.raw is not None and "." not in expr.raw:
            return INT
        return FLOAT
    if isinstance(expr, Str):
        return STRING
    if isinstance(expr, Nil):
        return NIL
    if isinstance(expr, Var):
        return env.lookup(expr.name)
    if isinstance(expr, Unary):
        inner = _infer_expr_type(expr.expr, env)
        if expr.op == "!":
            return BOOL
        if inner in {INT, FLOAT}:
            return inner
        return ANY
    if isinstance(expr, Bin):
        left = _infer_expr_type(expr.a, env)
        right = _infer_expr_type(expr.b, env)
        if expr.op in {"==", "!=", "<", ">", "<=", ">=", "&&", "||"}:
            return BOOL
        if expr.op == "+" and (left == STRING or right == STRING):
            return STRING
        if expr.op in {"+", "-", "*", "/"}:
            return combine_types(left, right)
    if isinstance(expr, ListLit):
        return LIST
    if isinstance(expr, (MapLit, RecordLiteral, WorkflowDef, GoalDef)):
        return RECORD
    if isinstance(expr, (Index, Attr)):
        return ANY
    if isinstance(expr, IndexAssign):
        return _infer_expr_type(expr.value, env)
    if isinstance(expr, Assign):
        return _infer_expr_type(expr.expr, env)
    if isinstance(expr, FnExpr):
        params = [_name_to_type(param.type_hint) for param in expr.params]
        return FunctionType(params, _name_to_type(expr.return_type))
    if isinstance(expr, Call):
        callee = _infer_expr_type(expr.callee, env)
        if isinstance(callee, FunctionType):
            return callee.return_type
    return ANY


class _DocumentIndexer:
    def __init__(self, *, server: "LanguageServer", path: str, uri: str, text: str, tokens: list[Tok], ast: list):
        self.server = server
        self.path = path
        self.uri = uri
        self.text = text
        self.lines = text.splitlines() or [""]
        self.tokens = tokens
        self.ast = ast
        self.definitions: list[DefinitionRecord] = []
        self.references: list[ReferenceRecord] = []
        self.visible_symbols: dict[str, DefinitionRecord] = {}
        self.module_aliases: dict[str, ModuleRecord] = {}
        self.scope_stack: list[dict[str, DefinitionRecord]] = [{}]
        self.type_env = _TypeEnv()
        self.import_state = {"loaded": set(), "loading": set(), "exports": {}, "modules": {}, "module_ids": {}}
        ensure_project_root(self.import_state, os.path.dirname(self.path), self.path)

    def build(self) -> DocumentState:
        self._predeclare(self.ast)
        for stmt in self.ast:
            self._walk_stmt(stmt)
        return DocumentState(
            uri=self.uri,
            path=self.path,
            text=self.text,
            diagnostics=[],
            tokens=self.tokens,
            definitions=self.definitions,
            references=self.references,
            visible_symbols=self.visible_symbols,
            module_aliases=self.module_aliases,
        )

    def _current_scope(self) -> dict[str, DefinitionRecord]:
        return self.scope_stack[-1]

    def _push_scope(self) -> None:
        self.scope_stack.append({})
        self.type_env.push()

    def _pop_scope(self) -> None:
        self.scope_stack.pop()
        self.type_env.pop()

    def _bind(self, definition: DefinitionRecord) -> None:
        self._current_scope()[definition.name] = definition
        self.visible_symbols[definition.name] = definition
        if definition.type_text is not None:
            self.type_env.bind(definition.name, _name_to_type(definition.type_text))

    def _lookup(self, name: str) -> DefinitionRecord | None:
        for scope in reversed(self.scope_stack):
            if name in scope:
                return scope[name]
        return None

    def _function_signature(self, stmt: FnDef) -> str:
        params = []
        for param in stmt.params:
            params.append(f"{param.name}: {param.type_hint}" if param.type_hint else param.name)
        suffix = f" -> {stmt.return_type}" if stmt.return_type else ""
        return f"fn {stmt.name}({', '.join(params)}){suffix}"

    def _add_definition(
        self,
        name: str,
        kind: str,
        line: int,
        col: int,
        detail: str,
        *,
        type_text: str | None = None,
        signature: str | None = None,
    ) -> DefinitionRecord:
        definition = DefinitionRecord(
            name=name,
            kind=kind,
            path=self.path,
            uri=self.uri,
            line=line,
            col=col,
            end_col=col + len(name),
            detail=detail,
            type_text=type_text,
            signature=signature,
        )
        self.definitions.append(definition)
        self._bind(definition)
        return definition

    def _add_reference(self, name: str, line: int, col: int, target: DefinitionRecord | None) -> None:
        self.references.append(
            ReferenceRecord(name=name, line=line, col=col, end_col=col + len(name), target=target)
        )

    def _predeclare(self, stmts: list) -> None:
        for stmt in stmts:
            if isinstance(stmt, FnDef):
                tok = getattr(stmt, "_tok", None)
                line = tok.line if tok is not None else 1
                col = _identifier_column(self.lines, line, stmt.name, (tok.col + 2) if tok is not None else 1)
                signature = self._function_signature(stmt)
                self._add_definition(stmt.name, "function", line, col, signature, type_text="function", signature=signature)
                self.type_env.bind(
                    stmt.name,
                    FunctionType([_name_to_type(param.type_hint) for param in stmt.params], _name_to_type(stmt.return_type)),
                )
            elif isinstance(stmt, (WorkflowDef, GoalDef)):
                tok = getattr(stmt, "_tok", None)
                line = tok.line if tok is not None else 1
                col = _identifier_column(self.lines, line, stmt.name, (tok.col + 1) if tok is not None else 1)
                self._add_definition(stmt.name, "variable", line, col, stmt.name, type_text="record")

    def _load_module(self, import_path: str, tok: Tok | None) -> ModuleRecord | None:
        try:
            full_path = resolve_import_path(import_path, os.path.dirname(self.path), self.import_state, tok, self.path)
        except Exception:
            return None
        return self.server.load_module_record(full_path, self.import_state)

    def _walk_stmt(self, stmt) -> None:
        if isinstance(stmt, Import):
            tok = getattr(stmt, "_tok", None)
            module_record = self._load_module(stmt.path, tok)
            if stmt.alias is not None and tok is not None and module_record is not None:
                col = _identifier_column(self.lines, tok.line, stmt.alias, tok.col)
                definition = self._add_definition(
                    stmt.alias,
                    "module",
                    tok.line,
                    col,
                    f"module {Path(module_record.path).stem}",
                )
                self.module_aliases[stmt.alias] = module_record
                self._current_scope()[stmt.alias] = definition
            if module_record is not None:
                names = stmt.names if stmt.names else (None if stmt.alias is not None else list(module_record.exports))
                if names is not None:
                    for name in names:
                        exported = module_record.exports.get(name)
                        if exported is not None:
                            self._current_scope()[name] = exported
                            self.visible_symbols[name] = exported
                            if exported.type_text is not None:
                                self.type_env.bind(name, _name_to_type(exported.type_text))
            return

        if isinstance(stmt, Let):
            self._walk_expr(stmt.expr)
            tok = getattr(stmt, "_tok", None)
            line = tok.line if tok is not None else 1
            col = _identifier_column(self.lines, line, stmt.name, (tok.col + 3) if tok is not None else 1)
            type_text = stmt.type_hint or _type_to_text(_infer_expr_type(stmt.expr, self.type_env))
            self._add_definition(stmt.name, "variable", line, col, f"let {stmt.name}", type_text=type_text)
            return

        if isinstance(stmt, FnDef):
            self._push_scope()
            for param in stmt.params:
                self._bind_param(param, stmt)
            self._walk_stmt(stmt.body)
            self._pop_scope()
            return

        if isinstance(stmt, Block):
            self._push_scope()
            for inner in stmt.stmts:
                self._walk_stmt(inner)
            self._pop_scope()
            return

        if isinstance(stmt, ExprStmt):
            self._walk_expr(stmt.expr)
            return

        if isinstance(stmt, Print):
            self._walk_expr(stmt.expr)
            return

        if isinstance(stmt, If):
            self._walk_expr(stmt.cond)
            self._walk_stmt(stmt.then_branch)
            if stmt.else_branch is not None:
                self._walk_stmt(stmt.else_branch)
            return

        if isinstance(stmt, While):
            self._walk_expr(stmt.cond)
            self._walk_stmt(stmt.body)
            return

        if isinstance(stmt, For):
            self._push_scope()
            if stmt.init is not None:
                self._walk_stmt(stmt.init)
            if stmt.cond is not None:
                self._walk_expr(stmt.cond)
            if stmt.inc is not None:
                self._walk_expr(stmt.inc)
            self._walk_stmt(stmt.body)
            self._pop_scope()
            return

        if isinstance(stmt, ForEach):
            self._walk_expr(stmt.iterable)
            self._push_scope()
            tok = getattr(stmt, "_tok", None)
            line = tok.line if tok is not None else 1
            col = _identifier_column(self.lines, line, stmt.name, (tok.col + 3) if tok is not None else 1)
            self._add_definition(stmt.name, "variable", line, col, f"for {stmt.name}", type_text="any")
            self._walk_stmt(stmt.body)
            self._pop_scope()
            return

        if isinstance(stmt, Return) and stmt.expr is not None:
            self._walk_expr(stmt.expr)
            return

        if isinstance(stmt, Yield) and stmt.expr is not None:
            self._walk_expr(stmt.expr)
            return

        if isinstance(stmt, WorkflowStateDecl):
            self._walk_expr(stmt.value)
            return

        if isinstance(stmt, TryCatch):
            self._walk_stmt(stmt.try_block)
            self._push_scope()
            tok = getattr(stmt, "_tok", None)
            line = tok.line if tok is not None else 1
            col = _identifier_column(self.lines, line, stmt.catch_var, tok.col if tok is not None else 1)
            self._add_definition(stmt.catch_var, "variable", line, col, f"catch {stmt.catch_var}", type_text="string")
            self._walk_stmt(stmt.catch_block)
            self._pop_scope()
            if stmt.finally_block is not None:
                self._walk_stmt(stmt.finally_block)
            return

        if isinstance(stmt, Throw):
            self._walk_expr(stmt.expr)

    def _bind_param(self, param: Param, fn_stmt: FnDef) -> None:
        tok = getattr(param, "_tok", None)
        line = tok.line if tok is not None else getattr(getattr(fn_stmt, "_tok", None), "line", 1)
        col = tok.col if tok is not None else 1
        self._add_definition(param.name, "variable", line, col, f"param {param.name}", type_text=param.type_hint or "any")

    def _walk_expr(self, expr) -> None:
        if isinstance(expr, Var):
            tok = getattr(expr, "_tok", None)
            if tok is not None:
                self._add_reference(expr.name, tok.line, tok.col, self._lookup(expr.name))
            return

        if isinstance(expr, Assign):
            tok = getattr(expr, "_tok", None)
            if tok is not None:
                col = _identifier_column(self.lines, tok.line, expr.name, 1)
                self._add_reference(expr.name, tok.line, col, self._lookup(expr.name))
            self._walk_expr(expr.expr)
            return

        if isinstance(expr, Unary):
            self._walk_expr(expr.expr)
            return

        if isinstance(expr, Bin):
            self._walk_expr(expr.a)
            self._walk_expr(expr.b)
            return

        if isinstance(expr, ListLit):
            for item in expr.items:
                self._walk_expr(item)
            return

        if isinstance(expr, MapLit):
            for key, value in expr.items:
                self._walk_expr(key)
                self._walk_expr(value)
            return

        if isinstance(expr, RecordLiteral):
            for _key, value in expr.fields:
                self._walk_expr(value)
            return

        if isinstance(expr, Index):
            self._walk_expr(expr.seq)
            self._walk_expr(expr.index)
            return

        if isinstance(expr, IndexAssign):
            self._walk_expr(expr.seq)
            self._walk_expr(expr.index)
            self._walk_expr(expr.value)
            return

        if isinstance(expr, Attr):
            self._walk_expr(expr.obj)
            tok = getattr(expr, "_tok", None)
            if isinstance(expr.obj, Var) and tok is not None:
                module = self.module_aliases.get(expr.obj.name)
                if module is not None:
                    target = module.exports.get(expr.name)
                    col = _identifier_column(self.lines, tok.line, expr.name, tok.col + 1)
                    self._add_reference(expr.name, tok.line, col, target)
            return

        if isinstance(expr, Call):
            self._walk_expr(expr.callee)
            for arg in expr.args:
                self._walk_expr(arg)
            return

        if isinstance(expr, FnExpr):
            self._push_scope()
            for param in expr.params:
                tok = getattr(param, "_tok", None)
                line = tok.line if tok is not None else 1
                col = tok.col if tok is not None else 1
                self._add_definition(param.name, "variable", line, col, f"param {param.name}", type_text=param.type_hint or "any")
            self._walk_stmt(expr.body)
            self._pop_scope()


class LanguageServer:
    def __init__(self, input_stream: BinaryIO | None = None, output_stream: BinaryIO | None = None):
        self.input_stream = input_stream if input_stream is not None else sys.stdin.buffer
        self.output_stream = output_stream if output_stream is not None else sys.stdout.buffer
        self.documents: dict[str, DocumentState] = {}
        self.module_cache: dict[str, ModuleRecord] = {}
        self.current_diagnostics: dict[str, list[dict]] = {}
        self.dependency_graph: DependencyGraph | None = None
        self.diagnostic_engine = WorkspaceDiagnosticEngine()
        self.shutdown_requested = False
        self.exit_code = 0

    def run(self) -> int:
        while True:
            message = _read_message(self.input_stream)
            if message is None:
                break
            if self.handle_message(message):
                break
        return self.exit_code

    def handle_message(self, message: dict) -> bool:
        method = message.get("method")
        params = message.get("params", {})
        msg_id = message.get("id")
        if method == "exit":
            self.exit_code = 0 if self.shutdown_requested else 1
            return True
        try:
            if method == "initialize":
                self._send_response(msg_id, result=self._handle_initialize())
                return False
            if method == "shutdown":
                self.shutdown_requested = True
                self._send_response(msg_id, result=None)
                return False
            if method == "initialized":
                return False
            if method == "textDocument/didOpen":
                self._handle_did_open(params)
                return False
            if method == "textDocument/didChange":
                self._handle_did_change(params)
                return False
            if method == "textDocument/completion":
                self._send_response(msg_id, result=self._handle_completion(params))
                return False
            if method == "textDocument/hover":
                self._send_response(msg_id, result=self._handle_hover(params))
                return False
            if method == "textDocument/definition":
                self._send_response(msg_id, result=self._handle_definition(params))
                return False
            if msg_id is not None:
                self._send_response(msg_id, error={"code": -32601, "message": f"Method not found: {method}"})
            return False
        except Exception as err:
            if msg_id is not None:
                self._send_response(msg_id, error={"code": -32603, "message": str(err)})
            return False

    def _send(self, payload: dict) -> None:
        _write_message(self.output_stream, payload)

    def _send_response(self, msg_id, *, result=None, error=None) -> None:
        payload = {"jsonrpc": "2.0", "id": msg_id}
        if error is not None:
            payload["error"] = error
        else:
            payload["result"] = result
        self._send(payload)

    def _send_notification(self, method: str, params: dict) -> None:
        self._send({"jsonrpc": "2.0", "method": method, "params": params})

    def _handle_initialize(self) -> dict:
        return {
            "serverInfo": {"name": "nodus-lsp", "version": "0.9.0"},
            "capabilities": {
                "textDocumentSync": 1,
                "completionProvider": {"triggerCharacters": ["."]},
                "hoverProvider": True,
                "definitionProvider": True,
            },
        }

    def _handle_did_open(self, params: dict) -> None:
        text_document = params.get("textDocument", {})
        uri = text_document.get("uri")
        text = text_document.get("text", "")
        if not uri:
            return
        self.module_cache.clear()
        state = self._analyze_document(uri, text)
        self.documents[uri] = state
        self._refresh_workspace_diagnostics(state.path, text)

    def _handle_did_change(self, params: dict) -> None:
        text_document = params.get("textDocument", {})
        uri = text_document.get("uri")
        if not uri:
            return
        changes = params.get("contentChanges", [])
        if not changes:
            return
        self.module_cache.clear()
        state = self._analyze_document(uri, changes[-1].get("text", ""))
        self.documents[uri] = state
        self._refresh_workspace_diagnostics(state.path, state.text)

    def _handle_completion(self, params: dict) -> dict:
        state = self._get_document(params)
        if state is None:
            return {"isIncomplete": False, "items": []}
        position = params.get("position", {})
        line = int(position.get("line", 0))
        character = int(position.get("character", 0))
        lines = state.text.splitlines()
        line_text = lines[line] if line < len(lines) else ""
        before = line_text[:character]
        member_match = re.search(r"([A-Za-z_][A-Za-z0-9_]*)\.$", before)
        if member_match:
            alias = member_match.group(1)
            module = state.module_aliases.get(alias)
            if module is not None:
                items = [self._completion_item(item) for item in sorted(module.exports.values(), key=lambda value: value.name)]
                return {"isIncomplete": False, "items": items}
        prefix_match = IDENTIFIER_RE.search(before)
        prefix = prefix_match.group(0) if prefix_match else ""
        items: list[dict] = []
        seen: set[str] = set()
        for keyword in sorted(KEYWORDS):
            if not prefix or keyword.startswith(prefix):
                items.append({"label": keyword, "kind": COMPLETION_KIND_KEYWORD, "detail": "keyword"})
                seen.add(keyword)
        for definition in sorted(state.visible_symbols.values(), key=lambda item: (item.name, item.line, item.col)):
            if definition.name in seen:
                continue
            if prefix and not definition.name.startswith(prefix):
                continue
            items.append(self._completion_item(definition))
            seen.add(definition.name)
        return {"isIncomplete": False, "items": items}

    def _completion_item(self, definition: DefinitionRecord) -> dict:
        kind = {
            "function": COMPLETION_KIND_FUNCTION,
            "variable": COMPLETION_KIND_VARIABLE,
            "module": COMPLETION_KIND_MODULE,
        }.get(definition.kind, COMPLETION_KIND_TEXT)
        detail = definition.signature or definition.type_text or definition.detail
        return {"label": definition.name, "kind": kind, "detail": detail}

    def _handle_hover(self, params: dict) -> dict | None:
        state = self._get_document(params)
        if state is None:
            return None
        token = self._token_at_position(state, params.get("position", {}))
        if token is None or token.kind != "ID":
            return None
        target = self._resolve_symbol_at(state, token)
        if target is None:
            return None
        if target.kind == "module":
            value = target.detail
        elif target.signature:
            value = target.signature
        elif target.type_text:
            value = f"{target.name}: {target.type_text}"
        else:
            value = target.detail
        return {"contents": {"kind": "plaintext", "value": value}}

    def _handle_definition(self, params: dict):
        state = self._get_document(params)
        if state is None:
            return None
        token = self._token_at_position(state, params.get("position", {}))
        if token is None or token.kind != "ID":
            return None
        target = self._resolve_symbol_at(state, token)
        if target is None:
            return None
        return {"uri": target.uri, "range": _range_for(target.line, target.col, target.end_col)}

    def _resolve_symbol_at(self, state: DocumentState, token: Tok) -> DefinitionRecord | None:
        previous = self._previous_non_comment_token(state.tokens, token)
        if previous is not None and previous.kind == ".":
            alias_token = self._previous_non_comment_token(state.tokens, previous)
            if alias_token is not None and alias_token.kind == "ID":
                module = state.module_aliases.get(alias_token.val)
                if module is not None:
                    return module.exports.get(token.val)
        if token.val in state.module_aliases:
            module = state.module_aliases[token.val]
            return DefinitionRecord(
                name=token.val,
                kind="module",
                path=module.path,
                uri=module.uri,
                line=1,
                col=1,
                end_col=1 + len(token.val),
                detail=f"module {Path(module.path).stem}",
            )
        for reference in reversed(state.references):
            if reference.name == token.val and reference.line == token.line and reference.col == token.col:
                return reference.target
        for definition in state.definitions:
            if definition.name == token.val and definition.line == token.line and definition.col == token.col:
                return definition
        return state.visible_symbols.get(token.val)

    def _publish_diagnostics(self, path: str, diagnostics: list[dict]) -> None:
        normalized = os.path.realpath(path)
        uri = None
        for state in self.documents.values():
            if os.path.realpath(state.path) == normalized:
                uri = state.uri
                break
        if uri is None:
            uri = _path_to_uri(path)
        self._send_notification("textDocument/publishDiagnostics", {"uri": uri, "diagnostics": diagnostics})

    def _token_at_position(self, state: DocumentState, position: dict) -> Tok | None:
        line = int(position.get("line", 0)) + 1
        character = int(position.get("character", 0)) + 1
        for token in state.tokens:
            if token.line != line:
                continue
            end_col = token.col + len(token.val)
            if token.col <= character <= end_col:
                return token
        return None

    def _previous_non_comment_token(self, tokens: list[Tok], current: Tok) -> Tok | None:
        previous = None
        for token in tokens:
            if token is current:
                return previous
            if token.kind not in {"COMMENT", "SEP"}:
                previous = token
        return None

    def _get_document(self, params: dict) -> DocumentState | None:
        text_document = params.get("textDocument", {})
        uri = text_document.get("uri")
        if not uri:
            return None
        return self.documents.get(uri)

    def _analyze_document(self, uri: str, text: str) -> DocumentState:
        path = _uri_to_path(uri)
        tokens: list[Tok] = []
        try:
            tokens = tokenize(text)
            ast = Parser(tokens).parse()
        except Exception as err:
            return DocumentState(uri, path, text, [], tokens, [], [], {}, {})
        document = _DocumentIndexer(server=self, path=path, uri=uri, text=text, tokens=tokens, ast=ast).build()
        return document

    def _refresh_workspace_diagnostics(self, path: str, text: str) -> None:
        overlays = {state.path: state.text for state in self.documents.values()}
        affected = self._collect_affected_paths(path)
        merged: dict[str, list[dict]] = {}
        graph = self.dependency_graph
        updated_paths: set[str] = set()
        for affected_path in sorted(affected):
            result = self.diagnostic_engine.analyze(
                affected_path,
                source=overlays.get(affected_path, text if os.path.abspath(affected_path) == os.path.abspath(path) else None),
                overlays=overlays,
                affected_paths={affected_path},
                dependency_graph=graph,
            )
            graph = result.dependency_graph
            updated_paths.update(result.diagnostics_by_file)
            for file_path, diagnostics in result.diagnostics_by_file.items():
                merged[file_path] = [diagnostic.to_lsp() for diagnostic in diagnostics]
        self.dependency_graph = graph
        stale_paths = {existing for existing in self.current_diagnostics if existing in affected and existing not in updated_paths}
        for stale in sorted(stale_paths):
            self._publish_diagnostics(stale, [])
            self.current_diagnostics.pop(stale, None)
        for file_path, diagnostics in sorted(merged.items()):
            self._publish_diagnostics(file_path, diagnostics)
            self.current_diagnostics[file_path] = diagnostics

    def _collect_affected_paths(self, changed_path: str) -> set[str]:
        normalized = os.path.abspath(changed_path)
        affected = {normalized}
        if self.dependency_graph is None:
            return affected
        reverse: dict[str, set[str]] = {}
        for module_path, node in self.dependency_graph.modules.items():
            for imported in node.imported_modules:
                reverse.setdefault(os.path.abspath(imported), set()).add(os.path.abspath(module_path))
        pending = [normalized]
        while pending:
            current = pending.pop()
            for dependent in reverse.get(current, ()):
                if dependent in affected:
                    continue
                affected.add(dependent)
                pending.append(dependent)
        return affected

    def load_module_record(self, path: str, import_state: dict | None = None) -> ModuleRecord | None:
        resolved = os.path.abspath(path)
        cached = self.module_cache.get(resolved)
        if cached is not None:
            return cached
        try:
            text = self._document_text_for_path(resolved)
            if text is None:
                with open(resolved, "r", encoding="utf-8") as handle:
                    text = handle.read()
            tokens = tokenize(text)
            ast = Parser(tokens).parse()
        except Exception:
            return None
        state = import_state or {"loaded": set(), "loading": set(), "exports": {}, "modules": {}, "module_ids": {}}
        state.setdefault("loaded", set())
        state.setdefault("loading", set())
        state.setdefault("exports", {})
        state.setdefault("modules", {})
        state.setdefault("module_ids", {})
        ensure_project_root(state, os.path.dirname(resolved), resolved)
        try:
            info = collect_module_info(ast, resolved, f"__lsp__{len(state['module_ids'])}__")
            exported_names = info.exports
        except Exception:
            exported_names = set()
        top_level: dict[str, object] = {}
        export_from: list[object] = []
        for stmt in ast:
            if isinstance(stmt, (Let, FnDef, WorkflowDef, GoalDef)):
                top_level[stmt.name] = stmt
            elif stmt.__class__.__name__ == "ExportFrom":
                export_from.append(stmt)
        exports: dict[str, DefinitionRecord] = {}
        lines = text.splitlines() or [""]
        for name in exported_names:
            stmt = top_level.get(name)
            if stmt is None:
                continue
            tok = getattr(stmt, "_tok", None)
            line = tok.line if tok is not None else 1
            col = _identifier_column(lines, line, name, tok.col if tok is not None else 1)
            if isinstance(stmt, FnDef):
                signature = _DocumentIndexer(server=self, path=resolved, uri=_path_to_uri(resolved), text=text, tokens=tokens, ast=ast)._function_signature(stmt)
                exports[name] = DefinitionRecord(
                    name=name,
                    kind="function",
                    path=resolved,
                    uri=_path_to_uri(resolved),
                    line=line,
                    col=col,
                    end_col=col + len(name),
                    detail=signature,
                    type_text="function",
                    signature=signature,
                )
            elif isinstance(stmt, Let):
                exports[name] = DefinitionRecord(
                    name=name,
                    kind="variable",
                    path=resolved,
                    uri=_path_to_uri(resolved),
                    line=line,
                    col=col,
                    end_col=col + len(name),
                    detail=f"let {name}",
                    type_text=stmt.type_hint or "any",
                )
            else:
                exports[name] = DefinitionRecord(
                    name=name,
                    kind="variable",
                    path=resolved,
                    uri=_path_to_uri(resolved),
                    line=line,
                    col=col,
                    end_col=col + len(name),
                    detail=name,
                    type_text="record",
                )
        for stmt in export_from:
            tok = getattr(stmt, "_tok", None)
            try:
                module_path = resolve_import_path(stmt.path, os.path.dirname(resolved), state, tok, resolved)
                module = self.load_module_record(module_path, state)
            except Exception:
                module = None
            if module is None:
                continue
            for name in stmt.names:
                if name in module.exports:
                    exports[name] = module.exports[name]
        module_record = ModuleRecord(path=resolved, uri=_path_to_uri(resolved), exports=exports)
        self.module_cache[resolved] = module_record
        return module_record

    def _document_text_for_path(self, path: str) -> str | None:
        normalized = os.path.abspath(path)
        for state in self.documents.values():
            if os.path.abspath(state.path) == normalized:
                return state.text
        return None


def run_stdio_server() -> int:
    return LanguageServer().run()


if __name__ == "__main__":
    raise SystemExit(run_stdio_server())



"""Nodus entrypoint and public API facade."""

# Internal module map (for contributors):
# - frontend/lexer.py: tokenization
# - frontend/ast/ast_nodes.py: AST definitions
# - frontend/parser.py: syntax parsing
# - compiler/compiler.py: bytecode compiler
# - vm/vm.py: stack VM runtime
# - tooling/loader.py: import resolution + module records
# - builtins/nodus_builtins.py: builtin registry/constants
# - runtime/diagnostics.py: error types + formatter
# - tooling/repl.py: REPL loop
# - cli/cli.py: CLI entrypoints

from nodus.support.version import VERSION, __version__
from nodus.compiler.compiler import Compiler, FunctionInfo
from nodus.runtime.diagnostics import LangRuntimeError, LangSyntaxError, format_error
from nodus.frontend.lexer import Tok, tokenize
from nodus.frontend.parser import Parser
from nodus.result import Result
# NodusRuntime is the primary public embedding API.
# Added to __all__ in v1.0 for discoverability.
from nodus.runtime.embedding import NodusRuntime



def resolve_imports(*args, **kwargs):
    """Resolve and inline imports for a Nodus source string.

    Delegates to ``nodus.tooling.loader.resolve_imports``.
    Prefer the ``ModuleLoader`` API for new code.
    """
    from nodus.tooling.loader import resolve_imports as _resolve_imports

    return _resolve_imports(*args, **kwargs)


def run_source(*args, **kwargs):
    """Compile and execute a Nodus source string, returning a ``Result`` dict.

    Convenience wrapper around ``nodus.tooling.loader.run_source``.
    For embedding use cases, prefer ``NodusRuntime`` (``nodus.runtime.embedding``)
    which provides sandbox controls, execution limits, and host function registration.
    """
    from nodus.tooling.loader import run_source as _run_source

    return _run_source(*args, **kwargs)


def main(argv=None):
    """CLI entry point — delegates to ``nodus.cli.cli.main``."""
    from nodus.cli.cli import main as _main

    return _main(argv)


def __getattr__(name):
    if name == "VM":
        from nodus.vm.vm import VM as _VM

        return _VM
    raise AttributeError(name)


__all__ = [
    "VERSION",
    "__version__",
    "main",
    "Compiler",
    "FunctionInfo",
    "LangRuntimeError",
    "LangSyntaxError",
    "format_error",
    "Tok",
    "tokenize",
    "resolve_imports",
    "run_source",
    "Parser",
    "VM",
    "Result",
    "NodusRuntime",
]


if __name__ == "__main__":
    raise SystemExit(main())



"""Workflow/goal AST/runtime lowering helpers."""

from __future__ import annotations

from nodus.frontend.ast.ast_nodes import (
    ActionStmt,
    Assign,
    Attr,
    Bin,
    Block,
    Bool,
    Call,
    CheckpointStmt,
    Comment,
    DestructureLet,
    ExprStmt,
    FieldAssign,
    FnDef,
    FnExpr,
    For,
    ForEach,
    GoalDef,
    GoalStep,
    If,
    Import,
    Index,
    IndexAssign,
    Let,
    ListLit,
    ListPattern,
    MapLit,
    Nil,
    Num,
    Param,
    Print,
    RecordLiteral,
    RecordPattern,
    Return,
    Str,
    Throw,
    TryCatch,
    Unary,
    Var,
    VarPattern,
    While,
    WorkflowDef,
    WorkflowStateDecl,
    WorkflowStep,
)
from nodus.orchestration.task_graph import TaskGraph, TaskNode


WORKFLOW_MARKER = "__workflow__"
GOAL_MARKER = "__goal__"
STEP_OPTION_KEYS = {
    "timeout_ms",
    "retries",
    "retry_delay_ms",
    "cache",
    "cache_key",
    "worker",
    "worker_timeout_ms",
}


def lower_workflow_ast(workflow: WorkflowDef) -> MapLit:
    return _lower_flow_ast(workflow, marker=WORKFLOW_MARKER, execution_kind="workflow")


def lower_goal_ast(goal: GoalDef) -> MapLit:
    return _lower_flow_ast(goal, marker=GOAL_MARKER, execution_kind="goal")


def _lower_flow_ast(flow, *, marker: str, execution_kind: str) -> MapLit:
    state_init = _lower_state_init(flow)
    state_names = [state.name for state in flow.states]
    items = [
        (Str(marker), Str(execution_kind)),
        (Str("name"), Str(flow.name)),
        (Str("execution_kind"), Str(execution_kind)),
        (
            Str("steps"),
            ListLit([_lower_step_ast(step, state_names) for step in flow.steps]),
        ),
    ]
    if state_init is not None:
        items.append((Str("state_init"), state_init))
    if state_names:
        items.append((Str("state_keys"), ListLit([Str(name) for name in state_names])))
    return MapLit(items)


def _lower_state_init(flow: WorkflowDef | GoalDef) -> FnExpr | None:
    if not flow.states:
        return None
    state_var = "__workflow_state"
    state_names = {state.name for state in flow.states}
    rewriter = _StateRewriter(state_names, state_var, initial_locals={state_var})
    stmts = [Let(state_var, MapLit([]))]
    for state in flow.states:
        expr = rewriter.rewrite_expr(state.value)
        assign = IndexAssign(Var(state_var), Str(state.name), expr)
        stmts.append(ExprStmt(assign))
    stmts.append(Return(Var(state_var)))
    return FnExpr([], Block(stmts), return_type=None)


def _lower_step_ast(step: WorkflowStep | GoalStep, state_names: list[str]) -> MapLit:
    state_var = "__workflow_state"
    body = step.body
    rewriter = _StateRewriter(set(state_names), state_var, initial_locals=set(step.deps) | ({state_var} if state_names else set()))
    rewritten_body = rewriter.rewrite_stmt(body)
    if state_names:
        prelude = Let(state_var, Call(Var("workflow_state"), []))
        body_stmts = rewritten_body.stmts if isinstance(rewritten_body, Block) else [rewritten_body]
        rewritten_body = Block([prelude] + body_stmts)
    body = _return_last_action(rewritten_body)
    return MapLit(
        [
            (Str("name"), Str(step.name)),
            (Str("deps"), ListLit([Str(dep) for dep in step.deps])),
            (Str("fn"), FnExpr([Param(dep) for dep in step.deps], body, return_type=None)),
            (Str("options"), step.options if step.options is not None else MapLit([])),
        ]
    )


def _return_last_action(body: object) -> Block:
    if not isinstance(body, Block) or not body.stmts:
        return body if isinstance(body, Block) else Block([body])
    last = body.stmts[-1]
    if isinstance(last, ExprStmt) and isinstance(last.expr, Call) and _is_action_builtin(last.expr):
        stmts = list(body.stmts[:-1]) + [Return(last.expr)]
        return _mark_from(Block(stmts), body)
    return body


def _is_action_builtin(expr: Call) -> bool:
    return isinstance(expr.callee, Var) and expr.callee.name in {
        "__action_tool",
        "__action_agent",
        "__action_memory_put",
        "__action_memory_get",
        "__action_emit",
    }


def runtime_flow_kind(value) -> str | None:
    if not isinstance(value, dict):
        return None
    if value.get(WORKFLOW_MARKER) == "workflow":
        return "workflow"
    if value.get(GOAL_MARKER) == "goal":
        return "goal"
    return None


def is_workflow_value(value) -> bool:
    return runtime_flow_kind(value) == "workflow"


def is_goal_value(value) -> bool:
    return runtime_flow_kind(value) == "goal"


def unwrap_runtime_value(value):
    if hasattr(value, "value"):
        return value.value
    return value


def _find_flow_value(globals_dict: dict[str, object], flow_name: str | None, *, kind: str):
    matches = {}
    for name, value in globals_dict.items():
        unwrapped = unwrap_runtime_value(value)
        if runtime_flow_kind(unwrapped) == kind:
            matches[name] = unwrapped
    if flow_name is not None:
        direct = matches.get(flow_name)
        if direct is not None:
            return direct
        for value in matches.values():
            if value.get("name") == flow_name:
                return value
        return None
    if len(matches) == 1:
        return next(iter(matches.values()))
    return None


def find_workflow_value(globals_dict: dict[str, object], workflow_name: str | None = None):
    return _find_flow_value(globals_dict, workflow_name, kind="workflow")


def find_goal_value(globals_dict: dict[str, object], goal_name: str | None = None):
    return _find_flow_value(globals_dict, goal_name, kind="goal")


def _flow_name_candidates(globals_dict: dict[str, object], *, kind: str) -> list[str]:
    names = []
    for name, value in globals_dict.items():
        unwrapped = unwrap_runtime_value(value)
        if runtime_flow_kind(unwrapped) == kind:
            flow_name = unwrapped.get("name")
            names.append(flow_name if isinstance(flow_name, str) else name)
    names.sort()
    return names


def workflow_name_candidates(globals_dict: dict[str, object]) -> list[str]:
    return _flow_name_candidates(globals_dict, kind="workflow")


def goal_name_candidates(globals_dict: dict[str, object]) -> list[str]:
    return _flow_name_candidates(globals_dict, kind="goal")


def workflow_to_graph(vm, workflow_value, *, init_state: bool = False, task_ids_by_step: dict[str, str] | None = None) -> TaskGraph:
    kind = runtime_flow_kind(workflow_value)
    if kind not in {"workflow", "goal"}:
        vm.runtime_error("type", "workflow value expected")
    name = workflow_value.get("name")
    if not isinstance(name, str) or not name:
        vm.runtime_error("type", "workflow name must be a non-empty string")
    steps = workflow_value.get("steps")
    if not isinstance(steps, list) or not steps:
        vm.runtime_error("type", "workflow must define at least one step")

    by_name: dict[str, TaskNode] = {}
    ordered: list[tuple[str, dict]] = []
    for step in steps:
        if not isinstance(step, dict):
            vm.runtime_error("type", "workflow steps must be maps")
        step_name = step.get("name")
        if not isinstance(step_name, str) or not step_name:
            vm.runtime_error("type", "workflow step name must be a non-empty string")
        if step_name in by_name:
            vm.runtime_error("type", f"Duplicate workflow step: {step_name}")
        ordered.append((step_name, step))
        by_name[step_name] = None  # type: ignore[assignment]

    tasks: list[TaskNode] = []
    resolved: dict[str, TaskNode] = {}
    step_to_task: dict[str, str] = {}
    for step_name, step in ordered:
        fn = step.get("fn")
        closure = vm.ensure_function(fn, f"workflow step '{step_name}'")
        expected_arity = len(step.get("deps", [])) if isinstance(step.get("deps", []), list) else None
        if expected_arity is not None and len(closure.function.params) != expected_arity:
            vm.runtime_error(
                "call",
                f"Workflow step '{step_name}' expects {expected_arity} dependency input(s) but defines {len(closure.function.params)} parameter(s)",
            )
        options = step.get("options", {})
        if options is None:
            options = {}
        if not isinstance(options, dict):
            vm.runtime_error("type", f"Workflow step '{step_name}' options must be a map")
        task_id = None
        if isinstance(task_ids_by_step, dict):
            preserved_task_id = task_ids_by_step.get(step_name)
            if isinstance(preserved_task_id, str) and preserved_task_id:
                task_id = preserved_task_id
        if task_id is None:
            vm._task_counter += 1
            task_id = f"task_{vm._task_counter}"
        task = TaskNode(
            task_id=task_id,
            function=closure,
            dependencies=[],
            timeout_ms=_number_option(vm, options, "timeout_ms", step_name),
            max_retries=int(_number_option(vm, options, "retries", step_name, default=0) or 0),
            retry_delay_ms=float(_number_option(vm, options, "retry_delay_ms", step_name, default=0.0) or 0.0),
            cache=bool(options.get("cache", False)),
            cache_key=options.get("cache_key"),
            worker=_string_option(vm, options, "worker", step_name),
            worker_timeout_ms=_number_option(vm, options, "worker_timeout_ms", step_name),
            step_name=step_name,
        )
        tasks.append(task)
        resolved[step_name] = task
        step_to_task[step_name] = task_id

    for step_name, step in ordered:
        deps = step.get("deps", [])
        if not isinstance(deps, list):
            vm.runtime_error("type", f"Workflow step '{step_name}' deps must be a list")
        dep_nodes = []
        for dep in deps:
            if not isinstance(dep, str):
                vm.runtime_error("type", f"Workflow step '{step_name}' dependency names must be strings")
            dep_task = resolved.get(dep)
            if dep_task is None:
                vm.runtime_error("runtime", f"Workflow step '{step_name}' references unknown dependency '{dep}'")
            dep_nodes.append(dep_task)
        resolved[step_name].dependencies = dep_nodes

    metadata = {
        "workflow_name": name,
        "execution_kind": kind,
        "step_to_task": step_to_task,
        "task_to_step": {task_id: step for step, task_id in step_to_task.items()},
        "workflow_source_path": getattr(vm, "source_path", None),
        "workflow_source_code": getattr(vm, "source_code", None),
    }
    if kind == "goal":
        metadata["goal_name"] = name
    if init_state:
        state_init = workflow_value.get("state_init")
        if state_init is not None:
            closure = vm.ensure_function(state_init, "workflow state initializer")
            state = vm.run_closure(closure, [])
            if not isinstance(state, dict):
                vm.runtime_error("type", "workflow state initializer must return a map")
            metadata["workflow_state"] = state
        else:
            metadata["workflow_state"] = {}
        metadata["checkpoints"] = []

    return TaskGraph(tasks, metadata=metadata)


def _number_option(vm, options: dict, key: str, step_name: str, default=None):
    value = options.get(key, default)
    if value is None:
        return None
    return vm.ensure_number(value, f"workflow step '{step_name}' option {key}")


def _string_option(vm, options: dict, key: str, step_name: str):
    value = options.get(key)
    if value is None:
        return None
    if not isinstance(value, str):
        vm.runtime_error("type", f"workflow step '{step_name}' option {key} expects a string")
    return value


def _mark_from(node, original):
    tok = getattr(original, "_tok", None)
    if tok is not None:
        setattr(node, "_tok", tok)
    return node


def _collect_pattern_names(pattern) -> list[str]:
    names: list[str] = []
    if isinstance(pattern, VarPattern):
        names.append(pattern.name)
    elif isinstance(pattern, ListPattern):
        for item in pattern.elements:
            names.extend(_collect_pattern_names(item))
    elif isinstance(pattern, RecordPattern):
        for _key, value in pattern.fields:
            names.extend(_collect_pattern_names(value))
    return names


def _lower_action_expr(expr: ActionStmt):
    target = Str(expr.target) if expr.target is not None else Nil()
    if expr.kind == "tool":
        return _mark_from(Call(Var("__action_tool"), [target, expr.payload if expr.payload is not None else MapLit([])]), expr)
    if expr.kind == "agent":
        return _mark_from(Call(Var("__action_agent"), [target, expr.payload if expr.payload is not None else MapLit([])]), expr)
    if expr.kind == "memory_put":
        return _mark_from(Call(Var("__action_memory_put"), [target, expr.payload if expr.payload is not None else Nil()]), expr)
    if expr.kind == "memory_get":
        return _mark_from(Call(Var("__action_memory_get"), [target]), expr)
    if expr.kind == "emit":
        return _mark_from(Call(Var("__action_emit"), [target, expr.payload if expr.payload is not None else MapLit([])]), expr)
    return expr


class _StateRewriter:
    """Rewrites workflow/goal step bodies to reference shared state via a map variable.

    **What it does:**
    Transforms AST nodes inside workflow and goal step bodies so that any
    reference to a workflow state variable (e.g. ``version`` declared with
    ``state version = "0.1.0"``) is replaced with an index expression into
    a shared state map (e.g. ``__state["version"]``).  Assignments to state
    variables similarly become index-assign expressions on the state map.

    **Why at compile time (workflow lowering), not at runtime:**
    Workflows and goals are lowered from their AST representation to
    ``MapLit`` nodes by ``lower_workflow_ast`` / ``lower_goal_ast`` before
    bytecode compilation.  This lowering phase runs inside the compiler's
    ``compile_stmt`` for ``WorkflowDef`` / ``GoalDef`` nodes.  Doing the
    rewrite at compile time means the bytecode emitted for each step function
    is already in the flat, state-map form — no runtime introspection or
    special VM opcodes are needed for state access.

    **Inputs:**
    - ``state_names`` — the set of state variable names declared in the
      workflow/goal (from ``WorkflowStateDecl`` nodes).
    - ``state_var`` — the name of the hidden local variable holding the
      state map (e.g. ``"__state"``).
    - ``initial_locals`` — names already in scope at the entry of the step
      body (used to avoid incorrectly rewriting shadowing locals).

    **Outputs:**
    A rewritten AST subtree where state variable references have been
    replaced with ``Index(Var(state_var), Str(name))`` expressions and
    state variable assignments have been replaced with
    ``IndexAssign(Var(state_var), Str(name), value)`` expressions.

    **Transformation rules (before → after):**

    Read access::

        version                    →  __state["version"]

    Write access (let / assign)::

        let version = "1.0"        →  let version = "1.0"  (first definition
                                       also writes to __state["version"])
        version = expr             →  __state["version"] = expr

    Scope shadowing::

        let version = ...          shadows version — further refs in that
                                   scope use the local, not the state map.

    Nested function bodies are treated as separate scopes; state references
    inside them are NOT rewritten (they would need an explicit capture of
    the state map to access it, which is not currently supported).
    """

    def __init__(self, state_names: set[str], state_var: str, initial_locals: set[str] | None = None):
        self.state_names = set(state_names)
        self.state_var = state_var
        self.scopes: list[set[str]] = [set(initial_locals or set())]

    def _is_local(self, name: str) -> bool:
        return any(name in scope for scope in self.scopes)

    def _define(self, name: str) -> None:
        self.scopes[-1].add(name)

    def _enter_scope(self) -> None:
        self.scopes.append(set())

    def _exit_scope(self) -> None:
        self.scopes.pop()

    def rewrite_stmt(self, stmt):
        if isinstance(stmt, Block):
            self._enter_scope()
            out = Block([self.rewrite_stmt(s) for s in stmt.stmts])
            self._exit_scope()
            return _mark_from(out, stmt)
        if isinstance(stmt, Comment):
            return stmt
        if isinstance(stmt, ExprStmt):
            return _mark_from(ExprStmt(self.rewrite_expr(stmt.expr)), stmt)
        if isinstance(stmt, Let):
            expr = self.rewrite_expr(stmt.expr)
            out = Let(stmt.name, expr, type_hint=stmt.type_hint, exported=stmt.exported)
            self._define(stmt.name)
            return _mark_from(out, stmt)
        if isinstance(stmt, DestructureLet):
            expr = self.rewrite_expr(stmt.expr)
            out = DestructureLet(stmt.pattern, expr)
            for name in _collect_pattern_names(stmt.pattern):
                self._define(name)
            return _mark_from(out, stmt)
        if isinstance(stmt, Print):
            return _mark_from(Print(self.rewrite_expr(stmt.expr)), stmt)
        if isinstance(stmt, If):
            cond = self.rewrite_expr(stmt.cond)
            then_branch = self.rewrite_stmt(stmt.then_branch)
            else_branch = self.rewrite_stmt(stmt.else_branch) if stmt.else_branch is not None else None
            return _mark_from(If(cond, then_branch, else_branch), stmt)
        if isinstance(stmt, While):
            return _mark_from(While(self.rewrite_expr(stmt.cond), self.rewrite_stmt(stmt.body)), stmt)
        if isinstance(stmt, For):
            self._enter_scope()
            init = self.rewrite_stmt(stmt.init) if stmt.init is not None else None
            cond = self.rewrite_expr(stmt.cond) if stmt.cond is not None else None
            inc = self.rewrite_expr(stmt.inc) if stmt.inc is not None else None
            body = self.rewrite_stmt(stmt.body)
            self._exit_scope()
            return _mark_from(For(init, cond, inc, body), stmt)
        if isinstance(stmt, ForEach):
            iterable = self.rewrite_expr(stmt.iterable)
            self._enter_scope()
            self._define(stmt.name)
            body = self.rewrite_stmt(stmt.body)
            self._exit_scope()
            return _mark_from(ForEach(stmt.name, iterable, body), stmt)
        if isinstance(stmt, Return):
            expr = self.rewrite_expr(stmt.expr) if stmt.expr is not None else None
            return _mark_from(Return(expr), stmt)
        if isinstance(stmt, TryCatch):
            try_block = self.rewrite_stmt(stmt.try_block)
            self._enter_scope()
            self._define(stmt.catch_var)
            catch_block = self.rewrite_stmt(stmt.catch_block)
            self._exit_scope()
            finally_block = self.rewrite_stmt(stmt.finally_block) if stmt.finally_block is not None else None
            return _mark_from(TryCatch(try_block, stmt.catch_var, catch_block, finally_block), stmt)
        if isinstance(stmt, Throw):
            return _mark_from(Throw(self.rewrite_expr(stmt.expr)), stmt)
        if isinstance(stmt, FnDef):
            self._define(stmt.name)
            self._enter_scope()
            for param in stmt.params:
                self._define(param.name)
            body = self.rewrite_stmt(stmt.body)
            self._exit_scope()
            return _mark_from(FnDef(stmt.name, stmt.params, body, return_type=stmt.return_type, exported=stmt.exported), stmt)
        if isinstance(stmt, Import):
            return stmt
        if isinstance(stmt, CheckpointStmt):
            return stmt
        return stmt

    def rewrite_expr(self, expr):
        if expr is None:
            return None
        if isinstance(expr, ActionStmt):
            lowered = _lower_action_expr(expr)
            return self.rewrite_expr(lowered)
        if isinstance(expr, (Num, Bool, Str, Nil)):
            return expr
        if isinstance(expr, Var):
            if expr.name in self.state_names and not self._is_local(expr.name):
                return _mark_from(Index(Var(self.state_var), Str(expr.name)), expr)
            return expr
        if isinstance(expr, Assign):
            value = self.rewrite_expr(expr.expr)
            if expr.name in self.state_names and not self._is_local(expr.name):
                return _mark_from(IndexAssign(Var(self.state_var), Str(expr.name), value), expr)
            return _mark_from(Assign(expr.name, value), expr)
        if isinstance(expr, Unary):
            return _mark_from(Unary(expr.op, self.rewrite_expr(expr.expr)), expr)
        if isinstance(expr, Bin):
            return _mark_from(Bin(expr.op, self.rewrite_expr(expr.a), self.rewrite_expr(expr.b)), expr)
        if isinstance(expr, ListLit):
            return _mark_from(ListLit([self.rewrite_expr(item) for item in expr.items]), expr)
        if isinstance(expr, MapLit):
            return _mark_from(MapLit([(self.rewrite_expr(k), self.rewrite_expr(v)) for k, v in expr.items]), expr)
        if isinstance(expr, RecordLiteral):
            return _mark_from(RecordLiteral([(key, self.rewrite_expr(value)) for key, value in expr.fields]), expr)
        if isinstance(expr, Index):
            return _mark_from(Index(self.rewrite_expr(expr.seq), self.rewrite_expr(expr.index)), expr)
        if isinstance(expr, IndexAssign):
            return _mark_from(IndexAssign(self.rewrite_expr(expr.seq), self.rewrite_expr(expr.index), self.rewrite_expr(expr.value)), expr)
        if isinstance(expr, Attr):
            return _mark_from(Attr(self.rewrite_expr(expr.obj), expr.name), expr)
        if isinstance(expr, FieldAssign):
            return _mark_from(FieldAssign(self.rewrite_expr(expr.obj), expr.name, self.rewrite_expr(expr.value)), expr)
        if isinstance(expr, Call):
            return _mark_from(Call(self.rewrite_expr(expr.callee), [self.rewrite_expr(arg) for arg in expr.args]), expr)
        if isinstance(expr, FnExpr):
            self._enter_scope()
            for param in expr.params:
                self._define(param.name)
            body = self.rewrite_stmt(expr.body)
            self._exit_scope()
            return _mark_from(FnExpr(expr.params, body, return_type=expr.return_type), expr)
        return expr



"""Channel runtime objects for coroutine communication."""

from __future__ import annotations

from collections import deque


class Channel:
    def __init__(self):
        self.queue = deque()
        self.waiting_receivers: deque[object] = deque()  # O(1) popleft
        self.waiting_senders: deque[tuple[object, object]] = deque()  # O(1) popleft
        self.closed: bool = False


class ChannelRecvRequest:
    def __init__(self, channel: Channel):
        self.channel = channel



"""Project manifest and lockfile helpers for Nodus tooling."""

from __future__ import annotations

import os
import tomllib
from dataclasses import dataclass


MANIFEST_NAME = "nodus.toml"
LOCKFILE_NAME = "nodus.lock"
NODUS_DIRNAME = ".nodus"
MODULES_DIRNAME = "modules"
SOURCE_DIRNAME = "src"
ENTRYPOINT_NAME = "main.nd"
CACHE_DIRNAME = "cache"


@dataclass(frozen=True)
class DependencySpec:
    name: str
    kind: str
    value: str


@dataclass(frozen=True)
class LockedPackage:
    name: str
    version: str
    source: str
    hash: str
    path: str | None = None


@dataclass
class ProjectConfig:
    root: str
    manifest_path: str
    lock_path: str
    nodus_dir: str
    modules_dir: str
    name: str
    version: str
    dependencies: dict[str, DependencySpec]
    registry_url: str | None = None


def find_project_root(start_dir: str) -> str | None:
    current = os.path.abspath(start_dir)
    while True:
        if os.path.isfile(os.path.join(current, MANIFEST_NAME)):
            return current
        parent = os.path.dirname(current)
        if parent == current:
            return None
        current = parent


def load_manifest(path: str) -> dict:
    with open(path, "rb") as handle:
        return tomllib.load(handle)


def parse_package(data: dict, *, root: str) -> tuple[str, str, str | None]:
    raw_package = data.get("package")
    if raw_package is not None:
        if not isinstance(raw_package, dict):
            raise ValueError("Manifest [package] must be a table")
        name = str(raw_package.get("name", os.path.basename(root)))
        version = str(raw_package.get("version", "0.1.0"))
        registry_url = str(raw_package.get("registry_url", "") or "") or None
        return name, version, registry_url
    return (
        str(data.get("name", os.path.basename(root))),
        str(data.get("version", "0.1.0")),
        None,
    )


def parse_dependencies(raw: dict) -> dict[str, DependencySpec]:
    if not isinstance(raw, dict):
        raise ValueError("Manifest [dependencies] must be a table")
    dependencies: dict[str, DependencySpec] = {}
    for raw_name, raw_value in raw.items():
        name = str(raw_name)
        if isinstance(raw_value, str):
            dependencies[name] = DependencySpec(name=name, kind="version", value=raw_value)
            continue
        if isinstance(raw_value, dict) and "path" in raw_value:
            dependencies[name] = DependencySpec(name=name, kind="path", value=str(raw_value["path"]))
            continue
        raise ValueError(f"Unsupported dependency spec for {name}")
    return dependencies


def write_project_manifest(
    path: str,
    *,
    name: str,
    version: str,
    dependencies: dict[str, DependencySpec],
    registry_url: str | None = None,
) -> None:
    lines = [
        "[package]",
        f'name = "{_escape(name)}"',
        f'version = "{_escape(version)}"',
    ]
    if registry_url:
        lines.append(f'registry_url = "{_escape(registry_url)}"')
    lines += [
        "",
        "[dependencies]",
    ]
    for dep_name in sorted(dependencies):
        spec = dependencies[dep_name]
        if spec.kind == "version":
            lines.append(f'{dep_name} = "{_escape(spec.value)}"')
            continue
        if spec.kind == "path":
            lines.append(f'{dep_name} = {{ path = "{_escape(spec.value)}" }}')
            continue
        raise ValueError(f"Unsupported dependency spec for {dep_name}")
    text = "\n".join(lines) + "\n"
    with open(path, "w", encoding="utf-8") as handle:
        handle.write(text)


def load_project(root: str) -> ProjectConfig:
    root = os.path.abspath(root)
    manifest_path = os.path.join(root, MANIFEST_NAME)
    data = load_manifest(manifest_path)
    name, version, registry_url = parse_package(data, root=root)
    dependencies = parse_dependencies(data.get("dependencies", {}))
    nodus_dir = os.path.join(root, NODUS_DIRNAME)
    return ProjectConfig(
        root=root,
        manifest_path=manifest_path,
        lock_path=os.path.join(root, LOCKFILE_NAME),
        nodus_dir=nodus_dir,
        modules_dir=os.path.join(nodus_dir, MODULES_DIRNAME),
        name=name,
        version=version,
        dependencies=dependencies,
        registry_url=registry_url,
    )


def load_project_from(start_dir: str) -> ProjectConfig | None:
    root = find_project_root(start_dir)
    if root is None:
        return None
    return load_project(root)


def create_project(root: str, name: str | None = None, version: str = "0.1.0") -> ProjectConfig:
    root = os.path.abspath(root)
    os.makedirs(os.path.join(root, NODUS_DIRNAME, MODULES_DIRNAME), exist_ok=True)
    os.makedirs(os.path.join(root, NODUS_DIRNAME, CACHE_DIRNAME), exist_ok=True)
    src_dir = os.path.join(root, SOURCE_DIRNAME)
    os.makedirs(src_dir, exist_ok=True)
    manifest_path = os.path.join(root, MANIFEST_NAME)
    if not os.path.exists(manifest_path):
        project_name = name or os.path.basename(root)
        write_project_manifest(
            manifest_path,
            name=project_name,
            version=version,
            dependencies={},
        )
    entry_path = os.path.join(src_dir, ENTRYPOINT_NAME)
    if not os.path.exists(entry_path):
        with open(entry_path, "w", encoding="utf-8") as handle:
            handle.write('print("hello from nodus")\n')
    return load_project(root)


def project_entry_path(project: ProjectConfig) -> str:
    return os.path.join(project.root, SOURCE_DIRNAME, ENTRYPOINT_NAME)


def read_lockfile(path: str) -> dict[str, LockedPackage]:
    if not os.path.isfile(path):
        return {}
    data = load_manifest(path)
    packages = data.get("package", [])
    if not isinstance(packages, list):
        raise ValueError("Lockfile [[package]] entries must be an array of tables")
    locked: dict[str, LockedPackage] = {}
    for entry in packages:
        if not isinstance(entry, dict):
            continue
        name = str(entry.get("name", "")).strip()
        if not name:
            continue
        locked[name] = LockedPackage(
            name=name,
            version=str(entry.get("version", "0.0.0")),
            source=str(entry.get("source", "")),
            hash=str(entry.get("hash", "")),
            path=str(entry["path"]) if "path" in entry else None,
        )
    return locked


def write_lockfile(path: str, packages: dict[str, LockedPackage]) -> None:
    lines: list[str] = []
    for name in sorted(packages):
        package = packages[name]
        lines.append("[[package]]")
        lines.append(f'name = "{_escape(package.name)}"')
        lines.append(f'version = "{_escape(package.version)}"')
        lines.append(f'source = "{_escape(package.source)}"')
        if package.path is not None:
            lines.append(f'path = "{_escape(package.path)}"')
        lines.append(f'hash = "{_escape(package.hash)}"')
        lines.append("")
    text = "\n".join(lines).rstrip()
    if text:
        text += "\n"
    with open(path, "w", encoding="utf-8") as handle:
        handle.write(text)


def dependency_source_path(project: ProjectConfig, dependency: str) -> str:
    return os.path.join(project.modules_dir, dependency)


def _escape(value: str) -> str:
    return value.replace("\\", "\\\\").replace('"', '\\"')



"""Package management entrypoints for Nodus tooling."""

import os
import tempfile
from pathlib import Path

from nodus.tooling.installer import install_project
from nodus.tooling.project import (
    DependencySpec,
    ProjectConfig,
    create_project,
    load_project,
    read_lockfile,
    write_project_manifest,
)
from nodus.tooling.registry import Registry
from nodus.tooling.resolver import resolve_project_dependencies


def get_registry_token(registry_url: str | None = None, cli_token: str | None = None) -> str | None:
    """
    Resolve a registry token using three-tier priority:
    1. cli_token (from --registry-token flag) — highest priority
    2. NODUS_REGISTRY_TOKEN environment variable
    3. ~/.nodus/config.toml (UserConfig)
    Returns None if no token is found.
    """
    if cli_token:
        return cli_token
    env_token = os.environ.get("NODUS_REGISTRY_TOKEN")
    if env_token:
        return env_token
    from nodus.tooling.user_config import UserConfig
    return UserConfig().get_registry_token(registry_url)


def ensure_project(root: str) -> ProjectConfig:
    manifest_path = os.path.join(os.path.abspath(root), "nodus.toml")
    if not os.path.isfile(manifest_path):
        raise FileNotFoundError(f"Project manifest not found: {manifest_path}")
    return load_project(root)


def init_project(root: str) -> ProjectConfig:
    return create_project(root)


def install_dependencies_for_project(
    root: str,
    *,
    update: bool = False,
    registry_url: str | None = None,
    cli_token: str | None = None,
) -> dict[str, str]:
    from nodus.tooling.registry_client import RegistryClient

    project = ensure_project(root)

    # Resolve registry URL: parameter > env var > project config
    resolved_url = (
        registry_url
        or os.environ.get("NODUS_REGISTRY_URL", "").strip() or None
        or project.registry_url
    )

    if resolved_url:
        token = get_registry_token(resolved_url, cli_token=cli_token)
        registry_client: RegistryClient | None = RegistryClient(resolved_url, token=token)
        registry = None
    else:
        registry_client = None
        registry = Registry.from_project_root(project.root)

    resolution = resolve_project_dependencies(
        project,
        update=update,
        registry=registry,
        registry_client=registry_client,
    )
    resolved = install_project(project, resolution)
    out: dict[str, str] = {}
    for name, dep in resolved.items():
        out[name] = dep.path if dep.path is not None else dep.source
    return out


def list_dependencies(root: str) -> list[tuple[str, str]]:
    project = ensure_project(root)
    lock = read_lockfile(project.lock_path)
    out: list[tuple[str, str]] = []
    for name in sorted(project.dependencies):
        status = lock.get(name)
        if status is None:
            status_text = "not installed"
        elif status.source == "path" and status.path is not None:
            status_text = f"path:{status.path}"
        else:
            status_text = status.source
        out.append((name, status_text))
    return out


def add_dependency(root: str, package_name: str) -> ProjectConfig:
    project = ensure_project(root)
    registry = Registry.from_project_root(project.root)
    versions = registry.available_versions(package_name)
    if not versions:
        raise ValueError(f"Dependency {package_name} is not available in the local registry")
    latest = versions[-1]
    dependencies = dict(project.dependencies)
    dependencies[package_name] = DependencySpec(name=package_name, kind="version", value=latest.version)
    write_project_manifest(
        project.manifest_path,
        name=project.name,
        version=project.version,
        dependencies=dependencies,
    )
    install_dependencies_for_project(project.root, update=True)
    return load_project(project.root)


def remove_dependency(root: str, package_name: str) -> ProjectConfig:
    project = ensure_project(root)
    if package_name not in project.dependencies:
        raise ValueError(f"Dependency not declared: {package_name}")
    dependencies = dict(project.dependencies)
    dependencies.pop(package_name, None)
    write_project_manifest(
        project.manifest_path,
        name=project.name,
        version=project.version,
        dependencies=dependencies,
    )
    install_dependencies_for_project(project.root, update=True)
    return load_project(project.root)


def publish_package_to_registry(
    project_dir: str,
    registry_url: str | None = None,
    cli_token: str | None = None,
) -> int:
    """
    Build and publish the current project to the registry.
    Returns 0 on success, 1 on failure.
    """
    from nodus.tooling.registry_client import RegistryClient, RegistryError, create_package_archive

    project = load_project(project_dir)

    # Resolve registry URL
    resolved_url = (
        registry_url
        or os.environ.get("NODUS_REGISTRY_URL")
        or project.registry_url
    )
    if not resolved_url:
        print(
            "Error: No registry configured. Use --registry or set "
            "registry_url in nodus.toml."
        )
        return 1

    # Resolve token
    token = get_registry_token(resolved_url, cli_token=cli_token)
    if not token:
        print(
            "Error: No registry token found. Run 'nodus login' or "
            "set NODUS_REGISTRY_TOKEN."
        )
        return 1

    name = project.name
    version = project.version
    if not name or not version:
        print("Error: nodus.toml must have [package] name and version.")
        return 1

    client = RegistryClient(resolved_url, token=token)
    tmp_dir = tempfile.mkdtemp(prefix="nodus_publish_")
    archive_path = Path(tmp_dir) / f"{name}-{version}.tar.gz"
    try:
        print(f"Building archive for {name}@{version}...")
        sha256 = create_package_archive(
            Path(project_dir), archive_path, name=name, version=version
        )
        print(f"Publishing to {resolved_url}...")
        result = client.publish_package(name, version, archive_path, sha256)
        published_url = result.get("url", resolved_url)
        print(f"Published {name}@{version} → {published_url}")
        return 0
    except RegistryError as err:
        print(f"Publish failed: {err}")
        return 1
    finally:
        import shutil
        shutil.rmtree(tmp_dir, ignore_errors=True)



import os

from nodus.tooling.project import load_project


def test_manifest_parsing(tmp_path):
    manifest = tmp_path / "nodus.toml"
    manifest.write_text(
        '\n'.join(
            [
                'name = "demo"',
                'version = "0.1.0"',
                "",
                "[dependencies]",
                'json = "^1.0"',
                'utils = { path = "../utils" }',
            ]
        )
        + "\n",
        encoding="utf-8",
    )
    project = load_project(str(tmp_path))
    assert project.name == "demo"
    assert project.dependencies["json"].kind == "version"
    assert project.dependencies["utils"].kind == "path"
    assert os.path.basename(project.manifest_path) == "nodus.toml"



import io
import json
import os
import task_graph
import tempfile
import time
import unittest
from contextlib import redirect_stderr, redirect_stdout
import http.client

import nodus as lang
from nodus.runtime.module_loader import ModuleLoader
from nodus.services.agent_runtime import AGENT_REGISTRY, register_agent, unregister_agent
from nodus.services.memory_runtime import GLOBAL_MEMORY_STORE
from nodus.services.server import run_in_thread


def run_program(src: str, source_path: str = "goal.nd"):
    _loader = ModuleLoader(project_root=None)
    code, functions, code_locs = _loader.compile_only(
        src,
        module_name=source_path,
    )
    vm = lang.VM(code, functions, code_locs=code_locs, source_path=source_path)
    out_buf = io.StringIO()
    err_buf = io.StringIO()
    with redirect_stdout(out_buf), redirect_stderr(err_buf):
        vm.run()
    return vm, out_buf.getvalue().splitlines(), err_buf.getvalue()


class GoalDslTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.server, cls.thread = run_in_thread("127.0.0.1", 0, allowed_paths=["."])
        cls.port = cls.server.server_address[1]
        time.sleep(0.05)

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()
        cls.server.server_close()

    def setUp(self):
        GLOBAL_MEMORY_STORE.load_snapshot({})
        AGENT_REGISTRY.clear()

    def tearDown(self):
        GLOBAL_MEMORY_STORE.load_snapshot({})
        AGENT_REGISTRY.clear()

    def request(self, method: str, path: str, payload: dict | None = None):
        conn = http.client.HTTPConnection("127.0.0.1", self.port, timeout=5)
        body = None
        headers = {}
        if payload is not None:
            body = json.dumps(payload)
            headers["Content-Type"] = "application/json"
        conn.request(method, path, body=body, headers=headers)
        resp = conn.getresponse()
        data = resp.read().decode("utf-8")
        conn.close()
        return resp.status, json.loads(data) if data else {}

    def test_linear_goal(self):
        src = """
goal demo {
    step a { return 1 }
    step b after a { return a + 1 }
}

let result = run_goal(demo)
print(result["goal"])
print(result["steps"]["a"])
print(result["steps"]["b"])
"""
        _vm, out, err = run_program(src)
        self.assertEqual(out, ["demo", "1.0", "2.0"])
        self.assertEqual(err.strip(), "")

    def test_tool_action_result_is_step_result(self):
        src = """
goal demo {
    step research {
        action tool "nodus_check" with {
            code: "print(1 + 1)",
            filename: "inline.nd"
        }
    }
}

let result = run_goal(demo)
print(result["steps"]["research"]["ok"])
print(result["steps"]["research"]["stage"])
"""
        _vm, out, _err = run_program(src)
        self.assertEqual(out, ["true", "check"])

    def test_agent_action_result_propagates(self):
        register_agent("summarize", lambda payload: {"summary": payload["input"]["text"]})
        src = """
goal publish_article {
    step research {
        return { "text": "condensed" }
    }

    step summarize after research {
        action agent "summarize" with {
            input: research
        }
    }
}

let result = run_goal(publish_article)
print(result["steps"]["summarize"]["result"]["summary"])
"""
        _vm, out, _err = run_program(src)
        self.assertEqual(out, ["condensed"])

    def test_memory_actions(self):
        src = """
goal demo {
    step store {
        action memory_put "article_summary" { "value": "done" }
    }

    step fetch after store {
        let x = action memory_get "article_summary"
        return x["value"]
    }
}

let result = run_goal(demo)
print(result["steps"]["store"]["value"])
print(result["steps"]["fetch"])
"""
        _vm, out, _err = run_program(src)
        self.assertEqual(out, ["done", "done"])

    def test_plan_goal_uses_step_names(self):
        src = """
goal demo {
    step a { return 1 }
    step b after a { return a + 1 }
    step c after a { return a * 2 }
}

let plan = plan_goal(demo)
print(plan["goal"])
print(len(plan["parallel_groups"][1]))
"""
        _vm, out, _err = run_program(src)
        self.assertEqual(out, ["demo", "2.0"])

    def test_resume_goal(self):
        src = """
goal demo {
    state x = 1

    step a {
        x = x + 1
        checkpoint "after_a"
        return x
    }

    step b after a {
        return x
    }
}

let result = run_goal(demo)
"""
        with tempfile.TemporaryDirectory() as td:
            path = os.path.join(td, "goal_resume.nd")
            with open(path, "w", encoding="utf-8") as f:
                f.write(src)
            vm, _out, _err = run_program(src, source_path=path)
            result = vm.globals["result"]
            if hasattr(result, "value"):
                result = result.value
            graph_id = result["graph_id"]
            task_graph._GRAPH_REGISTRY.pop(graph_id, None)
            task_graph._GRAPH_VMS.pop(graph_id, None)

            resumed_vm = lang.VM([], {}, code_locs=[], source_path=None)
            resumed = resumed_vm.builtin_resume_goal(graph_id, "after_a")
            self.assertEqual(resumed["goal"], "demo")
            self.assertEqual(resumed["steps"]["a"], 3)
            self.assertEqual(resumed["steps"]["b"], 3)
            self.assertEqual(resumed["state"]["x"], 3)

    def test_goal_runtime_events(self):
        src = """
goal demo {
    step a {
        action emit "step_started" with { source: "demo" }
        return 1
    }

    step b after a {
        action memory_put "x" a
    }
}

run_goal(demo)
"""
        vm, _out, _err = run_program(src)
        event_types = [event.type for event in vm.event_bus.events()]
        self.assertIn("goal_start", event_types)
        self.assertIn("goal_step_start", event_types)
        self.assertIn("goal_action_start", event_types)
        self.assertIn("goal_action_complete", event_types)
        self.assertIn("goal_complete", event_types)

    def test_goal_api_endpoints(self):
        register_agent("summarize", lambda payload: {"summary": payload["input"]["text"]})
        code = """
goal publish_article {
    step research {
        return { "text": "condensed" }
    }

    step summarize after research {
        action agent "summarize" with {
            input: research
        }
    }
}
"""
        status, payload = self.request("POST", "/goal/run", {"code": code, "filename": "inline.nd"})
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertEqual(payload["result"]["goal"], "publish_article")
        self.assertEqual(payload["result"]["steps"]["summarize"]["result"]["summary"], "condensed")
        graph_id = payload["result"]["graph_id"]

        status, payload = self.request("POST", "/goal/plan", {"code": code, "filename": "inline.nd"})
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertEqual(payload["plan"]["goal"], "publish_article")
        self.assertEqual(payload["plan"]["parallel_groups"][1], ["summarize"])

        task_graph._GRAPH_REGISTRY.pop(graph_id, None)
        task_graph._GRAPH_VMS.pop(graph_id, None)
        status, payload = self.request("POST", "/goal/resume", {"graph_id": graph_id})
        self.assertEqual(status, 200)
        self.assertTrue(payload["ok"])
        self.assertEqual(payload["result"]["goal"], "publish_article")

    def test_goal_cli_commands(self):
        code = """
goal demo {
    step a { return 1 }
    step b after a { return a + 1 }
}
"""
        with tempfile.TemporaryDirectory() as td:
            path = os.path.join(td, "demo.nd")
            with open(path, "w", encoding="utf-8") as f:
                f.write(code)

            out = io.StringIO()
            with redirect_stdout(out):
                run_exit = lang.main(["nodus", "goal-run", path])
            self.assertEqual(run_exit, 0)
            run_payload = json.loads(out.getvalue().strip())
            self.assertEqual(run_payload["goal"], "demo")
            self.assertEqual(run_payload["steps"]["b"], 2)
            graph_id = run_payload["graph_id"]

            out = io.StringIO()
            with redirect_stdout(out):
                plan_exit = lang.main(["nodus", "goal-plan", path])
            self.assertEqual(plan_exit, 0)
            plan_payload = json.loads(out.getvalue().strip())
            self.assertEqual(plan_payload["parallel_groups"][1], ["b"])

            task_graph._GRAPH_REGISTRY.pop(graph_id, None)
            task_graph._GRAPH_VMS.pop(graph_id, None)
            out = io.StringIO()
            with redirect_stdout(out):
                resume_exit = lang.main(["nodus", "goal-resume", graph_id])
            self.assertEqual(resume_exit, 0)
            resume_payload = json.loads(out.getvalue().strip())
            self.assertEqual(resume_payload["steps"]["b"], 2)


if __name__ == "__main__":
    unittest.main()



import io
import json
import os
import tempfile
import time
import unittest

from nodus.dap.server import DebugAdapterServer


def _decode_messages(data: bytes) -> list[dict]:
    out: list[dict] = []
    stream = io.BytesIO(data)
    while True:
        header = stream.readline()
        if not header:
            break
        if not header.strip():
            continue
        length = int(header.decode("ascii").split(":", 1)[1].strip())
        blank = stream.readline()
        if blank not in {b"\r\n", b"\n"}:
            raise AssertionError("Invalid DAP framing")
        body = stream.read(length)
        out.append(json.loads(body.decode("utf-8")))
    return out


class _Client:
    def __init__(self):
        self.output = io.BytesIO()
        self.server = DebugAdapterServer(io.BytesIO(), self.output)
        self.offset = 0
        self.seq = 1

    def request(self, command: str, arguments: dict | None = None) -> list[dict]:
        self.server.handle_message(
            {
                "seq": self.seq,
                "type": "request",
                "command": command,
                "arguments": arguments or {},
            }
        )
        self.seq += 1
        return self.drain()

    def drain(self) -> list[dict]:
        data = self.output.getvalue()
        chunk = data[self.offset :]
        self.offset = len(data)
        return _decode_messages(chunk)


class DapServerTests(unittest.TestCase):
    def test_initialize_handshake(self):
        client = _Client()
        messages = client.request("initialize")
        response = [message for message in messages if message["type"] == "response"][0]
        events = [message for message in messages if message["type"] == "event"]
        self.assertTrue(response["success"])
        self.assertEqual(response["command"], "initialize")
        self.assertTrue(response["body"]["supportsPauseRequest"])
        self.assertTrue(any(message["event"] == "initialized" for message in events))

    def test_breakpoint_registration_and_stop(self):
        with tempfile.TemporaryDirectory() as td:
            path = os.path.join(td, "main.nd")
            with open(path, "w", encoding="utf-8") as handle:
                handle.write("let x = 1\nx = x + 1\nprint(x)\n")

            client = _Client()
            client.request("initialize")
            client.request("launch", {"program": path})
            client.request("setBreakpoints", {"source": {"path": path}, "lines": [2]})

            debugger = client.server.session.debugger
            assert debugger is not None
            previous = debugger.stop_count
            client.request("continue")
            self.assertTrue(debugger.wait_for_stop(previous, timeout=1.0))
            messages = client.drain()

            self.assertIn((os.path.abspath(path), 2), debugger.breakpoints)
            stopped = [message for message in messages if message["type"] == "event" and message["event"] == "stopped"]
            self.assertTrue(stopped)
            self.assertEqual(stopped[-1]["body"]["reason"], "breakpoint")

            client.request("disconnect")

    def test_step_execution_emits_step_stop(self):
        with tempfile.TemporaryDirectory() as td:
            path = os.path.join(td, "step.nd")
            with open(path, "w", encoding="utf-8") as handle:
                handle.write("let x = 1\nx = x + 1\nprint(x)\n")

            client = _Client()
            client.request("initialize")
            client.request("launch", {"program": path})

            debugger = client.server.session.debugger
            assert debugger is not None
            previous = debugger.stop_count
            client.request("stepIn")
            self.assertTrue(debugger.wait_for_stop(previous, timeout=1.0))
            messages = client.drain()

            stopped = [message for message in messages if message["type"] == "event" and message["event"] == "stopped"]
            self.assertTrue(stopped)
            self.assertEqual(stopped[-1]["body"]["reason"], "step")

            client.request("disconnect")

    def test_variable_inspection_exposes_arguments_and_locals(self):
        with tempfile.TemporaryDirectory() as td:
            path = os.path.join(td, "vars.nd")
            with open(path, "w", encoding="utf-8") as handle:
                handle.write(
                    "fn foo(a) {\n"
                    "    let x = a + 1\n"
                    "    print(x)\n"
                    "}\n"
                    "foo(4)\n"
                )

            client = _Client()
            client.request("initialize")
            client.request("launch", {"program": path})
            client.request("setBreakpoints", {"source": {"path": path}, "lines": [3]})

            debugger = client.server.session.debugger
            assert debugger is not None
            previous = debugger.stop_count
            client.request("continue")
            self.assertTrue(debugger.wait_for_stop(previous, timeout=1.0))
            client.drain()

            stack_messages = client.request("stackTrace", {})
            stack_response = [message for message in stack_messages if message["type"] == "response"][0]
            frame_id = stack_response["body"]["stackFrames"][0]["id"]
            self.assertEqual(stack_response["body"]["stackFrames"][0]["name"], "foo")

            scope_messages = client.request("scopes", {"frameId": frame_id})
            scope_response = [message for message in scope_messages if message["type"] == "response"][0]
            scopes = scope_response["body"]["scopes"]
            args_ref = [scope["variablesReference"] for scope in scopes if scope["name"] == "Arguments"][0]
            locals_ref = [scope["variablesReference"] for scope in scopes if scope["name"] == "Locals"][0]

            args_messages = client.request("variables", {"variablesReference": args_ref})
            args_response = [message for message in args_messages if message["type"] == "response"][0]
            local_messages = client.request("variables", {"variablesReference": locals_ref})
            local_response = [message for message in local_messages if message["type"] == "response"][0]

            args = {entry["name"]: entry["value"] for entry in args_response["body"]["variables"]}
            locals_ = {entry["name"]: entry["value"] for entry in local_response["body"]["variables"]}
            self.assertEqual(args["a"], "4.0")
            self.assertEqual(locals_["x"], "5.0")

            client.request("disconnect")


if __name__ == "__main__":
    unittest.main()



"""Tests for finally block support in try/catch/finally."""

import io
import unittest
from contextlib import redirect_stdout

import nodus as lang
from nodus.runtime.module_loader import ModuleLoader


def run_program(src: str) -> list[str]:
    loader = ModuleLoader(project_root=None)
    code, functions, code_locs = loader.compile_only(src, module_name="<test>")
    vm = lang.VM(code, functions, code_locs=code_locs)
    buf = io.StringIO()
    with redirect_stdout(buf):
        vm.run()
    return buf.getvalue().splitlines()


class FinallyNormalPathTests(unittest.TestCase):
    """Finally runs when the try block completes without error."""

    def test_finally_runs_on_no_exception(self):
        src = """
try {
    print("try")
} catch err {
    print("catch")
} finally {
    print("finally")
}
"""
        self.assertEqual(run_program(src), ["try", "finally"])

    def test_finally_runs_after_try_body(self):
        src = """
let x = 0
try {
    x = 1
} catch err {
    x = 99
} finally {
    x = x + 10
}
print(x)
"""
        self.assertEqual(run_program(src), ["11.0"])

    def test_code_after_finally_executes(self):
        src = """
try {
    print("a")
} catch err {
    print("b")
} finally {
    print("c")
}
print("d")
"""
        self.assertEqual(run_program(src), ["a", "c", "d"])


class FinallyCaughtExceptionTests(unittest.TestCase):
    """Finally runs after catch handles an exception."""

    def test_finally_runs_after_catch(self):
        src = """
try {
    throw "boom"
} catch err {
    print("caught")
} finally {
    print("finally")
}
"""
        self.assertEqual(run_program(src), ["caught", "finally"])

    def test_finally_after_catch_then_continues(self):
        src = """
try {
    throw "err"
} catch e {
    print("c")
} finally {
    print("f")
}
print("done")
"""
        self.assertEqual(run_program(src), ["c", "f", "done"])

    def test_finally_runs_after_thrown_error(self):
        src = """
try {
    throw "oops"
} catch e {
    print("caught")
} finally {
    print("clean")
}
"""
        self.assertEqual(run_program(src), ["caught", "clean"])

    def test_finally_runs_after_runtime_error_caught(self):
        src = """
try {
    let bad = 1 / 0
} catch err {
    print("div error caught")
} finally {
    print("cleanup done")
}
"""
        self.assertEqual(run_program(src), ["div error caught", "cleanup done"])


class FinallyReturnInsideTryTests(unittest.TestCase):
    """Finally runs when a return statement exits a try block early."""

    def test_finally_runs_on_return_from_try(self):
        src = """
fn go() {
    try {
        return 1
    } catch err {
        return 2
    } finally {
        print("finally")
    }
}
let r = go()
print(r)
"""
        self.assertEqual(run_program(src), ["finally", "1.0"])

    def test_finally_runs_before_function_exits(self):
        src = """
fn work() {
    try {
        print("try")
        return 42
    } catch err {
        print("catch")
    } finally {
        print("finally")
    }
}
let v = work()
print(v)
"""
        self.assertEqual(run_program(src), ["try", "finally", "42.0"])

    def test_deferred_return_value_is_preserved(self):
        src = """
fn compute() {
    try {
        return 100
    } catch err {
        return 0
    } finally {
        print("side effect")
    }
}
print(compute())
"""
        self.assertEqual(run_program(src), ["side effect", "100.0"])


class FinallyNestedTests(unittest.TestCase):
    """Nested try/catch/finally blocks."""

    def test_nested_finally_inner_first(self):
        src = """
try {
    try {
        print("inner try")
    } catch e {
        print("inner catch")
    } finally {
        print("inner finally")
    }
} catch err {
    print("outer catch")
} finally {
    print("outer finally")
}
"""
        self.assertEqual(run_program(src), ["inner try", "inner finally", "outer finally"])

    def test_nested_finally_with_exception(self):
        src = """
try {
    try {
        throw "inner"
    } catch e {
        print("inner caught")
    } finally {
        print("inner finally")
    }
} catch err {
    print("outer catch")
} finally {
    print("outer finally")
}
"""
        self.assertEqual(run_program(src), ["inner caught", "inner finally", "outer finally"])


class FinallyWithoutExceptionTests(unittest.TestCase):
    """Finally with complex control flow."""

    def test_finally_with_variable_mutation(self):
        src = """
let log = []
try {
    log = log + ["try"]
} catch err {
    log = log + ["catch"]
} finally {
    log = log + ["finally"]
}
print(log[0])
print(log[1])
"""
        self.assertEqual(run_program(src), ["try", "finally"])

    def test_try_catch_without_finally_unchanged(self):
        """Existing try/catch behavior is unaffected."""
        src = """
try {
    throw "x"
} catch e {
    print("caught")
}
print("after")
"""
        self.assertEqual(run_program(src), ["caught", "after"])

    def test_finally_in_function_multiple_calls(self):
        src = """
fn cleanup(x) {
    try {
        return x * 2
    } catch err {
        return 0
    } finally {
        print("done")
    }
}
print(cleanup(3))
print(cleanup(5))
"""
        self.assertEqual(run_program(src), ["done", "6.0", "done", "10.0"])



import os

from nodus.tooling.runner import run_source


def test_multiple_tasks_progress(tmp_path):
    script = """
let busy = coroutine(fn() {
    let i = 0
    while (i < 5000) {
        i = i + 1
    }
    print("busy-done")
})

let fast = coroutine(fn() {
    print("fast")
})

spawn(busy)
spawn(fast)
run_loop()
"""
    result, _vm = run_source(script, filename=str(tmp_path / "main.nd"))
    assert result.get("ok", False), result
    lines = result.get("stdout", "").splitlines()
    assert "fast" in lines
    assert "busy-done" in lines
    assert lines.index("fast") < lines.index("busy-done")


def test_long_running_task_rotates_with_budget(tmp_path):
    script = """
let heavy = coroutine(fn() {
    let i = 0
    while (i < 8000) {
        i = i + 1
    }
    print("heavy complete")
})

let quick = coroutine(fn() {
    print("quick")
})

spawn(heavy)
spawn(quick)
run_loop()
"""
    result, _vm = run_source(script, filename=str(tmp_path / "main.nd"))
    assert result.get("ok", False), result
    lines = result.get("stdout", "").splitlines()
    assert "quick" in lines
    assert "heavy complete" in lines
    assert lines.index("quick") < lines.index("heavy complete")



"""Runtime profiler for Nodus VM."""

from __future__ import annotations

import time


class Profiler:
    def __init__(self, enabled: bool = False) -> None:
        self.enabled = enabled
        self.opcode_counts: dict[str, int] = {}
        self.function_calls: dict[str, int] = {}
        self.function_time: dict[str, float] = {}
        self.start_time: float = 0.0
        self.end_time: float = 0.0
        self._call_stack: list[tuple[str, float]] = []

    def start(self) -> None:
        self.enabled = True
        self.opcode_counts = {}
        self.function_calls = {}
        self.function_time = {}
        self._call_stack = []
        self.start_time = time.perf_counter()
        self.end_time = 0.0

    def stop(self) -> None:
        if self.start_time:
            self.end_time = time.perf_counter()
        self.enabled = False

    def record_opcode(self, opcode: str) -> None:
        if not self.enabled:
            return
        self.opcode_counts[opcode] = self.opcode_counts.get(opcode, 0) + 1

    def record_function_call(self, name: str | None) -> None:
        if not self.enabled or not name:
            return
        self.function_calls[name] = self.function_calls.get(name, 0) + 1

    def enter_function(self, name: str | None) -> None:
        if not self.enabled or not name:
            return
        self._call_stack.append((name, time.perf_counter()))

    def exit_function(self, name: str | None) -> None:
        if not self.enabled or not self._call_stack:
            return
        active_name, start = self._call_stack.pop()
        elapsed = time.perf_counter() - start
        self.function_time[active_name] = self.function_time.get(active_name, 0.0) + elapsed

    def report(self) -> dict:
        if not self.start_time:
            total_ms = 0.0
        else:
            end = self.end_time or time.perf_counter()
            total_ms = max(0.0, (end - self.start_time) * 1000.0)

        names = set(self.function_calls) | set(self.function_time)
        functions = []
        for name in names:
            calls = self.function_calls.get(name, 0)
            time_ms = self.function_time.get(name, 0.0) * 1000.0
            functions.append({"name": name, "calls": calls, "time_ms": time_ms})
        functions.sort(key=lambda item: (-item["time_ms"], -item["calls"], item["name"]))

        return {
            "total_time_ms": total_ms,
            "opcode_counts": dict(self.opcode_counts),
            "functions": functions,
        }



"""Interactive debugger support for the Nodus VM."""

from __future__ import annotations

import os
from dataclasses import dataclass

from nodus.runtime.diagnostics import LangRuntimeError

STEP_IN = "step_in"
STEP_OVER = "step_over"
STEP_OUT = "step_out"
CONTINUE = "continue"


class DebuggerQuit(Exception):
    """Raised when the debugger requests VM termination."""


@dataclass
class PauseState:
    reason: str
    module: str | None
    line: int | None
    col: int | None
    function: str
    locals: dict | None


@dataclass
class DebuggerFrame:
    event: str
    module: str | None
    line: int | None
    col: int | None
    function: str
    locals: dict | None
    depth: int
    ip: int


@dataclass
class StackFrameInfo:
    function: str
    line: int | None
    module: str | None


def get_stack(vm) -> list[StackFrameInfo]:
    current_path, current_line, _current_col = vm.current_loc()
    entries: list[StackFrameInfo] = []
    if not vm.frames:
        entries.append(StackFrameInfo("main", current_line, current_path))
        return entries

    first_call = vm.frames[0]
    entries.append(StackFrameInfo("main", first_call.call_line, first_call.call_path or vm.source_path))
    for i in range(1, len(vm.frames)):
        caller = vm.display_name(vm.frames[i - 1].fn_name)
        entries.append(StackFrameInfo(caller, vm.frames[i].call_line, vm.frames[i].call_path))
    current_name = vm.display_name(vm.frames[-1].fn_name)
    entries.append(StackFrameInfo(current_name, current_line, current_path))
    return entries


def get_locals(vm) -> dict:
    locals_ = vm.current_locals()
    values = locals_ if locals_ is not None else vm.globals
    out: dict = {}
    for name, value in (values or {}).items():
        if hasattr(value, "value"):
            value = value.value
        out[name] = value
    # Also read slot-indexed locals from locals_array (LOAD_LOCAL_IDX path)
    if vm.frames:
        frame = vm.frames[-1]
        if frame.locals_array is not None and frame.locals_name_to_slot is not None:
            for name, slot in frame.locals_name_to_slot.items():
                arr_value = frame.locals_array[slot]
                if hasattr(arr_value, "value"):
                    arr_value = arr_value.value
                out[name] = arr_value
    return out


class Debugger:
    def __init__(self, input_fn=input, output_fn=print, start_paused: bool = False):
        self.input_fn = input_fn
        self.output_fn = output_fn
        self.start_paused = start_paused
        self.started = False
        self.breakpoints: set[tuple[str | None, int]] = set()
        self.mode = CONTINUE
        self.next_depth: int | None = None
        self.next_steps = 0
        self.skip_ip: int | None = None
        self.breakpoint_latch: tuple[str | None, int] | None = None
        self.last_pause: PauseState | None = None
        self.pause_requested = False
        self.terminate_requested = False

    def set_breakpoint(self, module: str | None, line: int) -> None:
        self.breakpoints.add((module, line))

    def clear_breakpoint(self, module: str | None, line: int) -> None:
        self.breakpoints.discard((module, line))

    def resume_continue(self, vm) -> None:
        self.mode = CONTINUE
        self.next_depth = None
        self.next_steps = 0
        self.skip_ip = vm.ip
        self.pause_requested = False

    def resume_step_in(self, vm) -> None:
        self.mode = STEP_IN
        self.next_depth = None
        self.next_steps = 0
        self.skip_ip = vm.ip
        self.pause_requested = False

    def resume_step_over(self, vm) -> None:
        self.mode = STEP_OVER
        self.next_depth = len(vm.frames)
        self.next_steps = 0
        self.skip_ip = vm.ip
        self.pause_requested = False

    def resume_step_out(self, vm) -> None:
        if not vm.frames:
            self.resume_continue(vm)
            return
        self.mode = STEP_OUT
        self.next_depth = max(len(vm.frames) - 1, 0)
        self.next_steps = 0
        self.skip_ip = vm.ip
        self.pause_requested = False

    def request_pause(self) -> None:
        self.pause_requested = True

    def request_terminate(self) -> None:
        self.terminate_requested = True

    def before_instruction(self, vm, instr: tuple) -> None:
        if self.terminate_requested:
            raise DebuggerQuit()
        if self.skip_ip is not None and vm.ip != self.skip_ip:
            self.skip_ip = None

        loc = vm.current_loc()
        module, line, col = loc
        if self.breakpoint_latch != (module, line):
            self.breakpoint_latch = None

        if not self.started:
            self.started = True
            if self.start_paused:
                frame = self._build_frame(vm, module, line, col, event="before")
                self._pause(vm, frame, "start")
                return

        if self.skip_ip == vm.ip:
            return

        frame = self._build_frame(vm, module, line, col, event="before")
        if self.pause_requested:
            self.pause_requested = False
            self._pause(vm, frame, "pause")
            return
        if self.should_pause(frame):
            self._pause(vm, frame, "breakpoint")

    def after_instruction(self, vm, instr: tuple) -> None:
        if self.terminate_requested:
            raise DebuggerQuit()
        module, line, col = vm.current_loc()
        frame = self._build_frame(vm, module, line, col, event="after")
        if self.should_pause(frame):
            if self.mode == STEP_IN:
                reason = "step"
            elif self.mode == STEP_OUT:
                reason = "stepOut"
            else:
                reason = "next"
            self._pause(vm, frame, reason)

    def should_pause(self, frame: DebuggerFrame) -> bool:
        if frame.event == "before":
            if frame.line is not None and (frame.module, frame.line) in self.breakpoints:
                if self.breakpoint_latch != (frame.module, frame.line):
                    self.breakpoint_latch = (frame.module, frame.line)
                    return True
            return False

        if self.mode == STEP_IN:
            return True

        if self.mode == STEP_OVER:
            self.next_steps += 1
            if self.next_depth is not None and self.next_steps >= 1 and frame.depth <= self.next_depth:
                return True
        if self.mode == STEP_OUT:
            self.next_steps += 1
            if self.next_depth is not None and self.next_steps >= 1 and frame.depth <= self.next_depth:
                return True

        return False

    def _build_frame(self, vm, module: str | None, line: int | None, col: int | None, *, event: str) -> DebuggerFrame:
        function = vm.display_name(vm.frames[-1].fn_name) if vm.frames else "main"
        locals_ = vm.current_locals()
        depth = len(vm.frames)
        return DebuggerFrame(
            event=event,
            module=module,
            line=line,
            col=col,
            function=function,
            locals=locals_,
            depth=depth,
            ip=vm.ip,
        )

    def _pause(self, vm, frame: DebuggerFrame, reason: str) -> None:
        locals_view = get_locals(vm)
        state = PauseState(
            reason=reason,
            module=frame.module,
            line=frame.line,
            col=frame.col,
            function=frame.function,
            locals=locals_view,
        )
        self.last_pause = state
        self.output_fn(self.format_pause(state))
        while True:
            raw = self.input_fn("(nodusdb) ")
            cmd = raw.strip()
            if not cmd:
                continue
            if cmd == "step":
                self.resume_step_in(vm)
                return
            if cmd == "next":
                self.resume_step_over(vm)
                return
            if cmd in {"out", "stepout"}:
                self.resume_step_out(vm)
                return
            if cmd in {"continue", "run"}:
                self.resume_continue(vm)
                return
            if cmd == "stack":
                for line_text in self.format_stack(vm):
                    self.output_fn(line_text)
                continue
            if cmd == "locals":
                for line_text in self.format_locals(vm):
                    self.output_fn(line_text)
                continue
            if cmd.startswith("print"):
                expr = cmd[len("print") :].strip()
                if not expr:
                    self.output_fn("Usage: print <variable>")
                    continue
                self.output_fn(self.format_value(vm, expr))
                continue
            if cmd.startswith("break"):
                if self._handle_break(vm, cmd):
                    continue
                self.output_fn("Usage: break <file>:<line> or break <line>")
                continue
            if cmd == "quit":
                raise DebuggerQuit()
            self.output_fn(f"Unknown debugger command: {cmd}")

    def _handle_break(self, vm, cmd: str) -> bool:
        parts = cmd.split(maxsplit=1)
        if len(parts) != 2:
            return False
        target = parts[1].strip()
        if target.isdigit():
            line_no = int(target)
            module = self._current_module(vm)
            if module is None:
                self.output_fn("No current module to bind breakpoint.")
                return True
            self.set_breakpoint(module, line_no)
            self.output_fn(f"Breakpoint set at {module}:{line_no}")
            return True
        file_part, line_no = self._parse_file_line(target)
        if file_part is None or line_no is None:
            return False
        module = self._resolve_module_path(vm, file_part)
        self.set_breakpoint(module, line_no)
        self.output_fn(f"Breakpoint set at {module}:{line_no}")
        return True

    def _parse_file_line(self, target: str) -> tuple[str | None, int | None]:
        if ":" not in target:
            return None, None
        file_text, line_text = target.rsplit(":", 1)
        if not line_text.isdigit():
            return None, None
        return file_text, int(line_text)

    def _current_module(self, vm) -> str | None:
        module, _line, _col = vm.current_loc()
        if module:
            return module
        return vm.source_path

    def _resolve_module_path(self, vm, file_text: str) -> str:
        if file_text.startswith("<") and file_text.endswith(">"):
            return file_text
        if os.path.isabs(file_text):
            return file_text
        base_path = self._current_module(vm)
        if base_path and os.path.isabs(base_path):
            base_dir = os.path.dirname(base_path)
            return os.path.abspath(os.path.join(base_dir, file_text))
        return os.path.abspath(file_text)

    def format_pause(self, state: PauseState) -> str:
        loc = self.format_location(state.module, state.line, state.col)
        return f"[debug] paused ({state.reason}) at {loc} in {state.function}()"

    def format_location(self, module: str | None, line: int | None, col: int | None) -> str:
        if module and line is not None and col is not None:
            return f"{module}:{line}:{col}"
        if module and line is not None:
            return f"{module}:{line}"
        if line is not None:
            return f"line {line}"
        if module:
            return module
        return "<unknown>"

    def format_stack(self, vm) -> list[str]:
        entries = get_stack(vm)
        out: list[str] = []
        for entry in entries:
            out.append(self.describe_frame(entry.function, entry.line, path=entry.module))
        return out

    def describe_frame(self, name: str, line: int | None, path: str | None = None) -> str:
        if path and line is not None:
            return f"{name}() line {line} ({path})"
        if line is not None:
            return f"{name}() line {line}"
        return f"{name}()"

    def format_locals(self, vm) -> list[str]:
        values = get_locals(vm)
        if not values:
            return ["<no locals>"]

        out: list[str] = []
        for name in sorted(values):
            value = values[name]
            out.append(f"{name} = {vm.value_to_string(value, quote_strings=True)}")
        return out

    def format_value(self, vm, name: str) -> str:
        try:
            value = self._resolve_value(vm, name)
        except LangRuntimeError as err:
            return f"<error> {err}"
        return vm.value_to_string(value, quote_strings=True)

    def _resolve_value(self, vm, name: str):
        locals_ = vm.current_locals()
        if locals_ is not None and name in locals_:
            value = locals_[name]
            if hasattr(value, "value"):
                return value.value
            return value
        # Check slot-indexed locals_array
        if vm.frames:
            frame = vm.frames[-1]
            if frame.locals_array is not None and frame.locals_name_to_slot is not None:
                slot = frame.locals_name_to_slot.get(name)
                if slot is not None:
                    arr_value = frame.locals_array[slot]
                    if hasattr(arr_value, "value"):
                        return arr_value.value
                    return arr_value
        if name in vm.globals:
            value = vm.globals[name]
            if hasattr(value, "value"):
                return value.value
            return value
        return vm.load_name(name)



from nodus.tooling.formatter import format_source

def test_format_foreach():
    src = """
for item in items {
total = total + item
}
"""
    out = format_source(src)

    assert "for item in items {" in out


"""Dependency graph resolution for Nodus tooling."""

from __future__ import annotations

import os
import shutil
from dataclasses import dataclass
from pathlib import Path

from nodus.tooling.project import DependencySpec, LockedPackage, ProjectConfig, load_project, read_lockfile
from nodus.tooling.registry import Registry, RegistryPackage
from nodus.tooling.semver import Version, VersionRange


@dataclass(frozen=True)
class ResolvedPackage:
    name: str
    version: str
    source: str
    path: str
    source_path: str | None
    dependencies: tuple[str, ...]


@dataclass
class ResolutionResult:
    project: ProjectConfig
    packages: dict[str, ResolvedPackage]
    graph: dict[str, tuple[str, ...]]
    install_order: list[str]


def resolve_project_dependencies(
    project: ProjectConfig,
    *,
    update: bool = False,
    registry: Registry | None = None,
    registry_client=None,  # RegistryClient | None
) -> ResolutionResult:
    registry = registry or Registry.from_project_root(project.root)
    locked = {} if update else read_lockfile(project.lock_path)
    packages: dict[str, ResolvedPackage] = {}
    graph: dict[str, tuple[str, ...]] = {}
    visiting: list[str] = []
    install_order: list[str] = []

    # Staging dir for downloaded registry packages; cleaned and recreated each run
    staging_dir: str | None = None
    if registry_client is not None:
        staging_dir = os.path.join(project.nodus_dir, "_staging")
        if os.path.isdir(staging_dir):
            shutil.rmtree(staging_dir)
        os.makedirs(staging_dir, exist_ok=True)

    def resolve_spec(owner_root: str, spec: DependencySpec) -> ResolvedPackage:
        if spec.name in packages:
            package = packages[spec.name]
            _validate_locked_or_resolved(spec, package.version)
            return package

        if spec.name in visiting:
            cycle = " -> ".join(visiting + [spec.name])
            raise ValueError(f"Cyclic dependency detected: {cycle}")

        visiting.append(spec.name)
        try:
            package = _resolve_single_dependency(
                owner_root, spec, locked, registry,
                registry_client=registry_client,
                staging_dir=staging_dir,
            )
            dependencies = _load_dependency_specs(package.path)
            child_names: list[str] = []
            for dep_spec in dependencies.values():
                child = resolve_spec(package.path, dep_spec)
                child_names.append(child.name)
            resolved = ResolvedPackage(
                name=package.name,
                version=package.version,
                source=package.source,
                path=package.path,
                source_path=package.source_path,
                dependencies=tuple(sorted(child_names)),
            )
            packages[resolved.name] = resolved
            graph[resolved.name] = resolved.dependencies
            install_order.append(resolved.name)
            return resolved
        finally:
            visiting.pop()

    for spec in project.dependencies.values():
        resolve_spec(project.root, spec)

    return ResolutionResult(
        project=project,
        packages=packages,
        graph=graph,
        install_order=install_order,
    )


def resolve_locked_packages(project: ProjectConfig) -> dict[str, LockedPackage]:
    return read_lockfile(project.lock_path)


def _resolve_single_dependency(
    owner_root: str,
    spec: DependencySpec,
    locked: dict[str, LockedPackage],
    registry: Registry,
    *,
    registry_client=None,  # RegistryClient | None
    staging_dir: str | None = None,
) -> ResolvedPackage:
    if spec.kind == "path":
        path = _resolve_path(owner_root, spec.value)
        dependency_project = load_project(path)
        return ResolvedPackage(
            name=spec.name,
            version=dependency_project.version,
            source="path",
            path=dependency_project.root,
            source_path=_normalize_source_value(owner_root, path),
            dependencies=tuple(),
        )

    if spec.kind == "version" and registry_client is not None and staging_dir is not None:
        locked_package = locked.get(spec.name)
        if locked_package is not None and locked_package.source == "registry":
            # Re-resolve against the locked version's exact constraint
            constraint = locked_package.version
        else:
            constraint = spec.value
        version_entry = registry_client.resolve_version(spec.name, constraint)
        registry_client.install_package(spec.name, version_entry, Path(staging_dir))
        pkg_path = os.path.join(staging_dir, spec.name)
        return ResolvedPackage(
            name=spec.name,
            version=version_entry["version"],
            source="registry",
            path=pkg_path,
            source_path=None,
            dependencies=tuple(),
        )

    locked_package = locked.get(spec.name)
    if locked_package is not None:
        _validate_locked_or_resolved(spec, locked_package.version)

    selected = _select_registry_package(spec, registry, locked_package)
    return ResolvedPackage(
        name=selected.name,
        version=selected.version,
        source=selected.source,
        path=selected.path,
        source_path=None,
        dependencies=tuple(),
    )


def _load_dependency_specs(path: str) -> dict[str, DependencySpec]:
    manifest_path = os.path.join(path, "nodus.toml")
    if not os.path.isfile(manifest_path):
        return {}
    return load_project(path).dependencies


def _select_registry_package(
    spec: DependencySpec,
    registry: Registry,
    locked_package: LockedPackage | None,
) -> RegistryPackage:
    candidates = registry.available_versions(spec.name)
    if not candidates:
        raise ValueError(f"Dependency {spec.name} is not available in the local registry")

    if locked_package is not None:
        locked = registry.get(spec.name, locked_package.version)
        if locked is None:
            raise ValueError(f"Locked dependency {spec.name}@{locked_package.version} is unavailable")
        _validate_locked_or_resolved(spec, locked.version)
        return locked

    matching = [candidate for candidate in candidates if _matches_requirement(spec.value, candidate.version)]
    if not matching:
        raise ValueError(f"No registry version for {spec.name} satisfies {spec.value}")
    matching.sort(key=lambda item: Version.parse(item.version), reverse=True)
    return matching[0]


def _resolve_path(owner_root: str, raw_path: str) -> str:
    if os.path.isabs(raw_path):
        return os.path.abspath(raw_path)
    return os.path.abspath(os.path.join(owner_root, raw_path))


def _normalize_source_value(owner_root: str, resolved_path: str) -> str:
    relative = os.path.relpath(resolved_path, owner_root)
    return relative.replace("\\", "/")


def _matches_requirement(requirement: str, version: str) -> bool:
    try:
        return VersionRange.parse(requirement).matches(Version.parse(version))
    except ValueError:
        return version == requirement


def _validate_locked_or_resolved(spec: DependencySpec, version: str) -> None:
    if spec.kind != "version":
        return
    if not _matches_requirement(spec.value, version):
        raise ValueError(f"Resolved version {version} does not satisfy {spec.name} {spec.value}")



"""Lexical scope tracking for the Nodus compiler.

The symbol table resolves variable names to one of three scope types:

``global``
    Defined at module top-level.  The VM stores and loads these through
    ``module_globals`` and the LOAD / STORE opcodes.

``local``
    Defined inside a function (any block nested inside ``fn``).  The VM
    stores and loads these through ``frame.locals`` and the STORE_ARG /
    LOAD_LOCAL / STORE opcodes.  The compiler emits the faster LOAD_LOCAL
    opcode (instead of LOAD) when it can confirm a symbol is local-scope and
    the access is inside a function body.

``upvalue``
    A variable defined in an enclosing function that is captured by a nested
    function (closure).  The compiler inserts Cell boxing for the captured
    variable in the enclosing frame and emits LOAD_UPVALUE / STORE_UPVALUE
    with an integer index into the closure's upvalue array.

Two-pass compilation design
----------------------------
Functions are compiled *before* the main module body.  Each function starts
its own Scope (kind="function") pushed with ``enter_scope("function")``.
Upvalue resolution (``resolve_upvalue``) can only proceed once all enclosing
function scopes are in the scope chain, so the two-pass order — inner functions
first, then the outer body — ensures that when the compiler processes a nested
``fn``, all enclosing scopes are already on the stack and ``resolve_upvalue``
can walk up to them.

This is also why ``_current_function_scope`` walks up the scope chain rather
than just looking at the top scope: block scopes (``if``, ``while``, ``for``)
are interleaved between function scopes.
"""

from dataclasses import dataclass


@dataclass
class Symbol:
    name: str
    scope: str  # "global", "local", "upvalue"
    index: int | None = None
    is_function: bool = False


@dataclass
class Upvalue:
    name: str
    is_local: bool
    index: int | None


class Scope:
    def __init__(self, parent=None, kind: str = "module"):
        self.parent = parent
        self.kind = kind
        self.symbols: dict[str, Symbol] = {}
        self.upvalues: list[Upvalue] = []
        self.local_slot_counter: int = 0  # counts local variable slots in this function scope
        self.all_local_slots: dict[str, int] = {}  # name → slot for ALL locals (any depth) in this function

    def define(self, name: str, scope: str, is_function: bool = False) -> Symbol:
        symbol = Symbol(name=name, scope=scope, is_function=is_function)
        self.symbols[name] = symbol
        return symbol


class SymbolTable:
    def __init__(self):
        self.current = Scope(kind="module")
        self.all_symbols: set[str] = set()

    def enter_scope(self, kind: str = "block") -> None:
        self.current = Scope(self.current, kind=kind)

    def exit_scope(self) -> None:
        if self.current.parent is not None:
            self.current = self.current.parent

    def define(self, name: str, is_function: bool = False) -> Symbol:
        scope_kind = "global" if self.current.kind == "module" else "local"
        symbol = self.current.define(name, scope_kind, is_function=is_function)
        self.all_symbols.add(name)
        # Assign a slot index to local variables inside function scopes
        if scope_kind == "local":
            func_scope = self._current_function_scope()
            if func_scope is not None:
                symbol.index = func_scope.local_slot_counter
                func_scope.all_local_slots[name] = symbol.index
                func_scope.local_slot_counter += 1
        return symbol

    def define_function(self, name: str) -> Symbol:
        return self.define(name, is_function=True)

    def is_defined_anywhere(self, name: str) -> bool:
        return name in self.all_symbols

    def is_defined_in_module(self, name: str) -> bool:
        scope = self.current
        while scope.parent is not None:
            scope = scope.parent
        return name in scope.symbols

    def resolve(self, name: str) -> Symbol | None:
        symbol = self.resolve_local(name)
        if symbol is not None:
            return symbol
        return self.resolve_upvalue(name)

    def resolve_local(self, name: str) -> Symbol | None:
        scope = self.current
        while scope:
            if name in scope.symbols:
                return scope.symbols[name]
            if scope.kind in {"function", "module"}:
                break
            scope = scope.parent
        return None

    def resolve_upvalue(self, name: str) -> Symbol | None:
        """Resolve ``name`` as an upvalue from an enclosing function scope.

        Algorithm (implemented by ``_resolve_upvalue_in``):
        1. Find the immediately enclosing function scope (``enclosing``).
        2. Walk scopes between the current function scope and ``enclosing``,
           looking for ``name``.
        3. If ``name`` is found with ``scope == "global"``, return it as-is
           (globals are accessed directly, not via upvalue indirection).
        4. If ``name`` is found in ``enclosing`` as a local variable, add an
           upvalue entry with ``is_local=True``.  The VM will call
           ``capture_local(enclosing_frame, name)`` at closure-creation time
           to box the local into a Cell.
        5. If ``name`` is found in ``enclosing`` as an upvalue itself (i.e.
           it was already captured from a *further* enclosing scope), add an
           upvalue entry with ``is_local=False`` and copy the outer upvalue's
           index.  The VM will re-use the outer Cell.
        6. If not found in ``enclosing``, recurse outward (step 5 of the outer
           recursion handles the chaining).

        Cell boxing and mutability
        --------------------------
        ``is_local=True`` means "the captured variable lives directly in the
        immediately enclosing function's ``frame.locals``."  The VM boxes that
        local into a ``Cell`` the first time a closure captures it
        (``capture_local``), and every closure sharing that capture reads/writes
        the same Cell object, preserving shared-mutable semantics.

        ``is_local=False`` means "the captured variable was already itself an
        upvalue in the enclosing function."  The VM fetches the Cell from the
        enclosing closure's upvalue list by index rather than from ``locals``.

        Returns ``None`` if ``name`` cannot be found in any enclosing scope.
        """
        func_scope = self._current_function_scope()
        if func_scope is None:
            return None
        return self._resolve_upvalue_in(func_scope, name)

    @property
    def frame_size(self) -> int:
        """Number of local variable slots needed for the current function."""
        func_scope = self._current_function_scope()
        if func_scope is None:
            return 0
        return func_scope.local_slot_counter

    def current_function_upvalues(self) -> list[Upvalue]:
        """Return the ordered list of upvalues captured by the current function scope.

        Called by the compiler at the end of compiling a ``fn`` body, after all
        inner function bodies and variable references have been processed.  The
        compiler passes the returned list to ``FunctionInfo`` so that the VM knows
        which variables to capture at the MAKE_CLOSURE opcode.

        Each ``Upvalue`` in the list corresponds to one slot in the closure's
        upvalue array at runtime.  The ``index`` field of a Symbol with
        ``scope="upvalue"`` is its position in this list, and it is the same integer
        passed to LOAD_UPVALUE / STORE_UPVALUE at runtime.

        Returns an empty list for module-level code (where there is no enclosing
        function scope) and for functions that capture no variables.
        """
        func_scope = self._current_function_scope()
        if func_scope is None:
            return []
        return list(func_scope.upvalues)

    def _current_function_scope(self) -> Scope | None:
        scope = self.current
        while scope:
            if scope.kind == "function":
                return scope
            scope = scope.parent
        return None

    def _add_upvalue(self, func_scope: Scope, symbol: Symbol, is_local: bool) -> Symbol:
        if symbol.name in func_scope.symbols and func_scope.symbols[symbol.name].scope == "upvalue":
            return func_scope.symbols[symbol.name]

        source_index = symbol.index  # local slot when is_local=True; outer upvalue idx when is_local=False
        upvalue = Upvalue(name=symbol.name, is_local=is_local, index=source_index)
        func_scope.upvalues.append(upvalue)
        up_index = len(func_scope.upvalues) - 1
        up_symbol = Symbol(name=symbol.name, scope="upvalue", index=up_index)
        func_scope.symbols[symbol.name] = up_symbol
        self.all_symbols.add(symbol.name)
        return up_symbol

    def _resolve_upvalue_in(self, func_scope: Scope, name: str) -> Symbol | None:
        enclosing = self._enclosing_function_scope(func_scope)
        if enclosing is None:
            return None

        scope = func_scope.parent
        while scope:
            if name in scope.symbols:
                symbol = scope.symbols[name]
                if symbol.scope == "global":
                    return symbol
                if scope.kind == "function":
                    if symbol.scope == "upvalue":
                        return self._add_upvalue(func_scope, symbol, is_local=False)
                    return self._add_upvalue(func_scope, symbol, is_local=True)
                return self._add_upvalue(func_scope, symbol, is_local=True)
            if scope == enclosing:
                break
            scope = scope.parent

        outer_symbol = self._resolve_upvalue_in(enclosing, name)
        if outer_symbol is None:
            return None
        if outer_symbol.scope == "global":
            return outer_symbol
        return self._add_upvalue(func_scope, outer_symbol, is_local=False)

    def _enclosing_function_scope(self, func_scope: Scope) -> Scope | None:
        scope = func_scope.parent
        while scope:
            if scope.kind == "function":
                return scope
            scope = scope.parent
        return None



"""Coroutine, channel, and scheduler builtin functions for the Nodus VM."""

from nodus.runtime.coroutine import Coroutine
from nodus.runtime.channel import Channel, ChannelRecvRequest
from nodus.runtime.scheduler import SleepRequest


def register(vm, registry) -> None:
    """Register coroutine, channel, and sleep builtins onto the registry."""

    def builtin_coroutine_create(value):
        closure = vm.ensure_function(value, "coroutine(fn)")
        if len(closure.function.params) != 0:
            vm.runtime_error("call", "coroutine(fn) expects a zero-argument function")
        return Coroutine(closure)

    def builtin_coroutine_status(value):
        coroutine = vm.ensure_coroutine(value, "coroutine_status(coroutine)")
        return coroutine.state

    def builtin_coroutine_resume(value):
        from nodus.vm.vm import Frame
        coroutine = vm.ensure_coroutine(value, "resume(coroutine)")
        if coroutine.state == "finished":
            vm.runtime_error("runtime", "Cannot resume finished coroutine")
        if coroutine.state == "running":
            vm.runtime_error("runtime", "Cannot resume running coroutine")

        caller_context = vm.save_execution_context()
        try:
            if coroutine.state == "created":
                call_path, call_line, call_col = vm.current_loc()
                coroutine.stack = list(coroutine.initial_args or [])
                coroutine.frames = []
                coroutine.handler_stack = []
                coroutine.pending_iter_next = None
                coroutine.pending_get_iter = False
                vm.load_coroutine_context(coroutine)
                coroutine.state = "running"
                fn = coroutine.closure.function
                if vm.max_frames is not None and len(vm.frames) + 1 > vm.max_frames:
                    vm.runtime_error("sandbox", "Call stack overflow")
                coro_frame = Frame(
                    return_ip=None,
                    locals={},
                    fn_name=fn.name,
                    call_line=call_line,
                    call_col=call_col,
                    call_path=call_path,
                    closure=coroutine.closure,
                )
                if fn.local_slots:
                    coro_frame.locals_name_to_slot = fn.local_slots
                vm.frames.append(coro_frame)
                if vm.profiler is not None and vm.profiler.enabled:
                    vm.profiler.enter_function(vm.display_name(fn.name))
                vm.ip = fn.addr
            else:
                vm.load_coroutine_context(coroutine)
                coroutine.state = "running"

            try:
                status, result = vm.execute()
            except Exception:
                coroutine.state = "finished"
                coroutine.ip = None
                coroutine.stack = []
                coroutine.frames = []
                coroutine.handler_stack = []
                coroutine.pending_iter_next = None
                coroutine.pending_get_iter = False
                raise
            if status in {"yield", "return"}:
                if status == "return":
                    coroutine.last_result = result
                return result
            return None
        finally:
            vm.restore_execution_context(caller_context)

    def builtin_spawn(value):
        coroutine = vm.ensure_coroutine(value, "spawn(coroutine)")
        vm.scheduler.spawn(coroutine)
        return None

    def builtin_run_loop():
        vm.scheduler.run_loop()
        return None

    def builtin_sleep(value):
        if isinstance(value, bool) or not isinstance(value, (int, float)):
            vm.runtime_error("type", "sleep(ms) expects a number")
        ms = float(value)
        if ms < 0:
            ms = 0.0
        return SleepRequest(ms)

    def builtin_channel():
        return Channel()

    def builtin_send(channel, value):
        ch = vm.ensure_channel(channel, "send(channel, value)")
        if ch.closed:
            vm.runtime_error("runtime", "send on closed channel")
        sender_id = vm.current_coroutine.id if vm.current_coroutine is not None else None
        sender_name = vm.current_coroutine.name if vm.current_coroutine is not None else None
        if ch.waiting_receivers:
            receiver = ch.waiting_receivers.popleft()
            if receiver.stack:
                receiver.stack[-1] = value
            receiver.blocked_on = None
            receiver.blocked_reason = None
            vm.scheduler.schedule(receiver)
            vm.event_bus.emit_event(
                "channel_send",
                coroutine_id=sender_id,
                name=sender_name,
                data={"queue_size": float(len(ch.queue)), "waiting_receivers": float(len(ch.waiting_receivers))},
            )
            vm.event_bus.emit_event(
                "channel_recv",
                coroutine_id=receiver.id,
                name=receiver.name,
                data={"from_wait": True},
            )
            vm.event_bus.emit_event("channel_wake", coroutine_id=receiver.id, name=receiver.name)
            return None
        ch.queue.append(value)
        vm.event_bus.emit_event(
            "channel_send",
            coroutine_id=sender_id,
            name=sender_name,
            data={"queue_size": float(len(ch.queue)), "waiting_receivers": float(len(ch.waiting_receivers))},
        )
        return None

    def builtin_recv(channel):
        ch = vm.ensure_channel(channel, "recv(channel)")
        if ch.queue:
            value = ch.queue.popleft()
            vm.event_bus.emit_event(
                "channel_recv",
                coroutine_id=vm.current_coroutine.id if vm.current_coroutine is not None else None,
                name=vm.current_coroutine.name if vm.current_coroutine is not None else None,
                data={"from_queue": True, "queue_size": float(len(ch.queue))},
            )
            return value
        if ch.closed:
            vm.event_bus.emit_event(
                "channel_recv",
                coroutine_id=vm.current_coroutine.id if vm.current_coroutine is not None else None,
                name=vm.current_coroutine.name if vm.current_coroutine is not None else None,
                data={"closed": True},
            )
            return None
        if vm.current_coroutine is None:
            vm.runtime_error("runtime", "recv(channel) outside coroutine")
        coroutine = vm.current_coroutine
        coroutine.state = "suspended"
        coroutine.blocked_on = ch
        coroutine.blocked_reason = "channel_recv"
        vm.stack.append(None)
        vm.save_current_coroutine_state(vm.ip + 1)
        ch.waiting_receivers.append(coroutine)
        vm.event_bus.emit_event(
            "channel_block",
            coroutine_id=coroutine.id,
            name=coroutine.name,
            data={"operation": "recv"},
        )
        return ChannelRecvRequest(ch)

    def builtin_close(channel):
        ch = vm.ensure_channel(channel, "close(channel)")
        if ch.closed:
            return None
        ch.closed = True
        vm.event_bus.emit_event(
            "channel_close",
            coroutine_id=vm.current_coroutine.id if vm.current_coroutine is not None else None,
            name=vm.current_coroutine.name if vm.current_coroutine is not None else None,
            data={"waiting_receivers": float(len(ch.waiting_receivers))},
        )
        while ch.waiting_receivers:
            receiver = ch.waiting_receivers.popleft()
            if getattr(receiver, "state", None) != "suspended":
                continue
            if receiver.stack:
                receiver.stack[-1] = None
            receiver.blocked_on = None
            receiver.blocked_reason = None
            vm.scheduler.schedule(receiver)
            vm.event_bus.emit_event("channel_wake", coroutine_id=receiver.id, name=receiver.name)
            vm.event_bus.emit_event(
                "channel_recv",
                coroutine_id=receiver.id,
                name=receiver.name,
                data={"closed": True},
            )
        return None

    registry.add("coroutine", 1, builtin_coroutine_create)
    registry.add("resume", 1, builtin_coroutine_resume)
    registry.add("coroutine_status", 1, builtin_coroutine_status)
    registry.add("spawn", 1, builtin_spawn)
    registry.add("run_loop", 0, builtin_run_loop)
    registry.add("sleep", 1, builtin_sleep)
    registry.add("__sleep", 1, builtin_sleep)
    registry.add("channel", 0, builtin_channel)
    registry.add("send", 2, builtin_send)
    registry.add("recv", 1, builtin_recv)
    registry.add("close", 1, builtin_close)
